# Copyright 2021 Vincent Texier <vit@free.fr>
#
# This software is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
from __future__ import annotations

import abc
from typing import TYPE_CHECKING

from tikka.interfaces.adapters.repository.technical_committee_members import (
    TechnicalCommitteeMembersRepositoryInterface,
)
from tikka.interfaces.adapters.repository.technical_committee_proposals import (
    TechnicalCommitteeProposalsRepositoryInterface,
)

if TYPE_CHECKING:
    from tikka.interfaces.adapters.repository.accounts import (
        AccountsRepositoryInterface,
    )
    from tikka.interfaces.adapters.repository.authorities import (
        AuthoritiesRepositoryInterface,
    )
    from tikka.interfaces.adapters.repository.categories import (
        CategoriesRepositoryInterface,
    )
    from tikka.interfaces.adapters.repository.config import ConfigRepositoryInterface
    from tikka.interfaces.adapters.repository.currencies import (
        CurrenciesRepositoryInterface,
    )
    from tikka.interfaces.adapters.repository.currency import (
        CurrencyRepositoryInterface,
    )
    from tikka.interfaces.adapters.repository.db_client import DBClientInterface
    from tikka.interfaces.adapters.repository.file_wallets import (
        V1FileWalletsRepositoryInterface,
    )
    from tikka.interfaces.adapters.repository.identities import (
        IdentitiesRepositoryInterface,
    )
    from tikka.interfaces.adapters.repository.indexers import (
        IndexersRepositoryInterface,
    )
    from tikka.interfaces.adapters.repository.nodes import NodesRepositoryInterface
    from tikka.interfaces.adapters.repository.passwords import (
        PasswordsRepositoryInterface,
    )
    from tikka.interfaces.adapters.repository.preferences import (
        PreferencesRepositoryInterface,
    )
    from tikka.interfaces.adapters.repository.smiths import SmithsRepositoryInterface
    from tikka.interfaces.adapters.repository.transfers import (
        TransfersRepositoryInterface,
    )
    from tikka.interfaces.adapters.repository.wallets import WalletsRepositoryInterface


class RepositoryInterface(abc.ABC):
    """
    RepositoryInterface class
    """

    @property
    def db_client(self) -> DBClientInterface:
        """
        Return ConfigRepositoryInterface instance

        :return:
        """
        raise NotImplementedError

    @property
    def config(self) -> ConfigRepositoryInterface:
        """
        Return ConfigRepositoryInterface instance

        :return:
        """
        raise NotImplementedError

    @property
    def currencies(self) -> CurrenciesRepositoryInterface:
        """
        Return CurrenciesRepositoryInterface instance

        :return:
        """
        raise NotImplementedError

    @property
    def v1_file_wallets(self) -> V1FileWalletsRepositoryInterface:
        """
        Return V1FileWalletsRepositoryInterface instance

        :return:
        """
        raise NotImplementedError

    @property
    def accounts(self) -> AccountsRepositoryInterface:
        """
        Return AccountsRepositoryInterface instance

        :return:
        """
        raise NotImplementedError

    @property
    def authorities(self) -> AuthoritiesRepositoryInterface:
        """
        Return AuthoritiesRepositoryInterface instance

        :return:
        """
        raise NotImplementedError

    @property
    def categories(self) -> CategoriesRepositoryInterface:
        """
        Return CategoriesRepositoryInterface instance

        :return:
        """
        raise NotImplementedError

    @property
    def currency(self) -> CurrencyRepositoryInterface:
        """
        Return CurrencyRepositoryInterface instance

        :return:
        """
        raise NotImplementedError

    @property
    def identities(self) -> IdentitiesRepositoryInterface:
        """
        Return IdentitiesRepositoryInterface instance

        :return:
        """
        raise NotImplementedError

    @property
    def indexers(self) -> IndexersRepositoryInterface:
        """
        Return IndexersRepositoryInterface instance

        :return:
        """
        raise NotImplementedError

    @property
    def nodes(self) -> NodesRepositoryInterface:
        """
        Return NodesRepositoryInterface instance

        :return:
        """
        raise NotImplementedError

    @property
    def passwords(self) -> PasswordsRepositoryInterface:
        """
        Return PasswordsRepositoryInterface instance

        :return:
        """
        raise NotImplementedError

    @property
    def preferences(self) -> PreferencesRepositoryInterface:
        """
        Return PreferencesRepositoryInterface instance

        :return:
        """
        raise NotImplementedError

    @property
    def smiths(self) -> SmithsRepositoryInterface:
        """
        Return SmithsRepositoryInterface instance

        :return:
        """
        raise NotImplementedError

    @property
    def transfers(self) -> TransfersRepositoryInterface:
        """
        Return TransfersRepositoryInterface instance

        :return:
        """
        raise NotImplementedError

    @property
    def wallets(self) -> WalletsRepositoryInterface:
        """
        Return WalletsRepositoryInterface instance

        :return:
        """
        raise NotImplementedError

    @property
    def technical_committee_members(
        self,
    ) -> TechnicalCommitteeMembersRepositoryInterface:
        """
        Return TechnicalCommitteeMembersRepositoryInterface instance

        :return:
        """
        raise NotImplementedError

    @property
    def technical_committee_proposals(
        self,
    ) -> TechnicalCommitteeProposalsRepositoryInterface:
        """
        Return TechnicalCommitteeProposalsRepositoryInterface instance

        :return:
        """
        raise NotImplementedError



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'tikka/slots/pyqt/resources/gui/windows/welcome.ui'
#
# Created by: PyQt5 UI code generator 5.15.10
#
# WARNING: Any manual changes made to this file will be lost when pyuic5 is
# run again.  Do not edit this file unless you know what you are doing.


from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_WelcomeDialog(object):
    def setupUi(self, WelcomeDialog):
        WelcomeDialog.setObjectName("WelcomeDialog")
        WelcomeDialog.resize(505, 672)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(WelcomeDialog.sizePolicy().hasHeightForWidth())
        WelcomeDialog.setSizePolicy(sizePolicy)
        self.verticalLayout_2 = QtWidgets.QVBoxLayout(WelcomeDialog)
        self.verticalLayout_2.setObjectName("verticalLayout_2")
        self.label = QtWidgets.QLabel(WelcomeDialog)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.label.sizePolicy().hasHeightForWidth())
        self.label.setSizePolicy(sizePolicy)
        self.label.setPixmap(QtGui.QPixmap(":/icons/logo"))
        self.label.setAlignment(QtCore.Qt.AlignCenter)
        self.label.setWordWrap(False)
        self.label.setObjectName("label")
        self.verticalLayout_2.addWidget(self.label)
        self.label_3 = QtWidgets.QLabel(WelcomeDialog)
        self.label_3.setAlignment(QtCore.Qt.AlignCenter)
        self.label_3.setObjectName("label_3")
        self.verticalLayout_2.addWidget(self.label_3)
        self.verticalLayout = QtWidgets.QVBoxLayout()
        self.verticalLayout.setContentsMargins(-1, 20, -1, 20)
        self.verticalLayout.setObjectName("verticalLayout")
        self.importV2Button = QtWidgets.QPushButton(WelcomeDialog)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.importV2Button.sizePolicy().hasHeightForWidth())
        self.importV2Button.setSizePolicy(sizePolicy)
        self.importV2Button.setObjectName("importV2Button")
        self.verticalLayout.addWidget(self.importV2Button, 0, QtCore.Qt.AlignHCenter)
        self.importV1Button = QtWidgets.QPushButton(WelcomeDialog)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.importV1Button.sizePolicy().hasHeightForWidth())
        self.importV1Button.setSizePolicy(sizePolicy)
        self.importV1Button.setObjectName("importV1Button")
        self.verticalLayout.addWidget(self.importV1Button, 0, QtCore.Qt.AlignHCenter)
        self.createButton = QtWidgets.QPushButton(WelcomeDialog)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.createButton.sizePolicy().hasHeightForWidth())
        self.createButton.setSizePolicy(sizePolicy)
        self.createButton.setObjectName("createButton")
        self.verticalLayout.addWidget(self.createButton, 0, QtCore.Qt.AlignHCenter)
        self.verticalLayout_2.addLayout(self.verticalLayout)
        self.continueButton = QtWidgets.QPushButton(WelcomeDialog)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.continueButton.sizePolicy().hasHeightForWidth())
        self.continueButton.setSizePolicy(sizePolicy)
        self.continueButton.setObjectName("continueButton")
        self.verticalLayout_2.addWidget(self.continueButton, 0, QtCore.Qt.AlignHCenter)

        self.retranslateUi(WelcomeDialog)
        QtCore.QMetaObject.connectSlotsByName(WelcomeDialog)
        WelcomeDialog.setTabOrder(self.createButton, self.continueButton)

    def retranslateUi(self, WelcomeDialog):
        #  _translate = QtCore.QCoreApplication.translate
        WelcomeDialog.setWindowTitle(_("Welcome"))
        self.label_3.setText(_("Welcome in Tikka !\n"
"\n"
"To get started, you can import an existing account or create a new one."))
        self.importV2Button.setText(_("Import account with mnemonic"))
        self.importV1Button.setText(_("Import V1 account with credentials"))
        self.createButton.setText(_("Create a new account"))
        self.continueButton.setText(_("Continue"))
import tikka.slots.pyqt.resources.icons.index_rc



# Copyright 2021 Vincent Texier <vit@free.fr>
#
# This software is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
from __future__ import annotations

import abc
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from interfaces.adapters.network.connection import ConnectionInterface
    from interfaces.adapters.network.indexer.identities import (
        IndexerIdentitiesInterface,
    )
    from interfaces.adapters.network.indexer.transfers import IndexerTransfersInterface

    from tikka.domains.entities.indexer import Indexer


class NetworkIndexerInterface(abc.ABC):
    """
    NetworkIndexerInterface class
    """

    @property
    def connection(self) -> ConnectionInterface:
        """
        Return ConnectionsInterface instance

        :return:
        """
        raise NotImplementedError

    @abc.abstractmethod
    def get(self) -> Indexer:
        """
        Return the node instance with infos from connection

        :return:
        """
        raise NotImplementedError

    @property
    def identities(self) -> IndexerIdentitiesInterface:
        """
        Return IndexerIdentitiesInterface instance

        :return:
        """
        raise NotImplementedError

    @property
    def transfers(self) -> IndexerTransfersInterface:
        """
        Return IndexerTransfersInterface instance

        :return:
        """
        raise NotImplementedError


class NetworkIndexerException(Exception):
    """
    NetworkIndexerException class
    """



# Copyright 2021 Vincent Texier <vit@free.fr>
#
# This software is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import abc
from typing import Any, Optional

from tikka.interfaces.entities.Server import ServerInterface


class ConnectionInterface(abc.ABC):
    """
    ConnectionInterface class
    """

    url: str = ""
    client: Optional[Any] = None

    @abc.abstractmethod
    def connect(self, server: ServerInterface) -> None:
        """
        Start connection from server

        :param server: Node instance
        :return:
        """
        raise NotImplementedError

    @abc.abstractmethod
    def disconnect(self) -> None:
        """
        Close connection

        :return:
        """
        raise NotImplementedError

    @abc.abstractmethod
    def is_connected(self) -> bool:
        """
        Return True if connection is active or OK.

        :return:
        """
        raise NotImplementedError


class NetworkConnectionError(ConnectionError):
    """
    NetworkConnectionError class
    """

    def __init__(self):
        super().__init__("Network connection error")



# Copyright 2021 Vincent Texier <vit@free.fr>
#
# This software is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from typing import Any, List, Optional

from sql import Column, Flavor, Table

from tikka.domains.entities.node import Node
from tikka.interfaces.adapters.repository.db_repository import DBRepositoryInterface
from tikka.interfaces.adapters.repository.nodes import NodesRepositoryInterface

TABLE_NAME = "nodes"

# create sql table wrapper
sql_nodes_table = Table(TABLE_NAME)

SQL_COLUMNS = {
    NodesRepositoryInterface.COLUMN_URL: "url",
    NodesRepositoryInterface.COLUMN_PEER_ID: "peer_id",
    NodesRepositoryInterface.COLUMN_BLOCK: "block",
    NodesRepositoryInterface.COLUMN_SOFTWARE: "software",
    NodesRepositoryInterface.COLUMN_SOFTWARE_VERSION: "software_version",
    NodesRepositoryInterface.COLUMN_SESSION_KEYS: "session_keys",
    NodesRepositoryInterface.COLUMN_EPOCH_INDEX: "epoch_index",
    NodesRepositoryInterface.COLUMN_UNSAFE_API_EXPOSED: "unsafe_api_exposed",
}

DEFAULT_LIST_OFFSET = 0
DEFAULT_LIST_LIMIT = 1000


class DBNodesRepository(NodesRepositoryInterface, DBRepositoryInterface):
    """
    DBNodesRepository class
    """

    def list(
        self,
        offset: int = DEFAULT_LIST_OFFSET,
        limit: int = DEFAULT_LIST_LIMIT,
        sort_column: str = NodesRepositoryInterface.COLUMN_URL,
        sort_order: str = NodesRepositoryInterface.SORT_ORDER_ASCENDING,
    ) -> List[Node]:
        __doc__ = (  # pylint: disable=redefined-builtin, unused-variable
            NodesRepositoryInterface.list.__doc__
        )

        sql_columns = {
            NodesRepositoryInterface.COLUMN_URL: sql_nodes_table.url,
            NodesRepositoryInterface.COLUMN_BLOCK: sql_nodes_table.block,
            NodesRepositoryInterface.COLUMN_SOFTWARE: sql_nodes_table.software,
            NodesRepositoryInterface.COLUMN_PEER_ID: sql_nodes_table.peer_id,
            NodesRepositoryInterface.COLUMN_EPOCH_INDEX: sql_nodes_table.epoch_index,
            NodesRepositoryInterface.COLUMN_UNSAFE_API_EXPOSED: sql_nodes_table.unsafe_api_exposed,
            NodesRepositoryInterface.COLUMN_SESSION_KEYS: sql_nodes_table.session_keys,
            NodesRepositoryInterface.COLUMN_SOFTWARE_VERSION: sql_nodes_table.software_version,
        }
        # if sort column...
        if sort_column is not None:
            # set sort column
            sql_sort_colum: Column = sql_columns[sort_column]
            # create select query wrapper with order by
            sql_select = sql_nodes_table.select(
                order_by=sql_sort_colum.asc
                if sort_order == NodesRepositoryInterface.SORT_ORDER_ASCENDING
                else sql_sort_colum.desc,
                offset=offset,
                limit=limit,
            )
        else:
            #  create select query wrapper without order by
            sql_select = sql_nodes_table.select(offset=offset, limit=limit)

        # config sql with ? as param style
        Flavor.set(Flavor(paramstyle="qmark"))

        sql, args = tuple(sql_select)
        result_set = self.client.select(sql, args)
        list_ = []
        for row in result_set:
            list_.append(get_node_from_row(row))

        return list_

    def add(self, node: Node) -> None:
        __doc__ = (  # pylint: disable=redefined-builtin, unused-variable
            NodesRepositoryInterface.add.__doc__
        )
        # insert only non protected fields
        self.client.insert(
            TABLE_NAME,
            **get_fields_from_node(node),
        )

    def get(self, url: str) -> Optional[Node]:
        __doc__ = (  # pylint: disable=redefined-builtin, unused-variable
            NodesRepositoryInterface.list.__doc__
        )

        row = self.client.select_one(f"SELECT * FROM {TABLE_NAME} WHERE url=?", (url,))
        if row is None:
            return None

        return get_node_from_row(row)

    def update(self, node: Node) -> None:
        __doc__ = (  # pylint: disable=redefined-builtin, unused-variable
            NodesRepositoryInterface.update.__doc__
        )

        # update only non hidden fields
        self.client.update(
            TABLE_NAME,
            f"url='{node.url}'",
            **get_fields_from_node(node),
        )

    def delete(self, url: str) -> None:
        __doc__ = (  # pylint: disable=redefined-builtin, unused-variable
            NodesRepositoryInterface.delete.__doc__
        )

        self.client.delete(TABLE_NAME, url=url)

    def delete_all(self) -> None:
        __doc__ = (  # pylint: disable=redefined-builtin, unused-variable
            NodesRepositoryInterface.delete_all.__doc__
        )

        self.client.clear(TABLE_NAME)

    def count(self) -> int:
        __doc__ = (  # pylint: disable=redefined-builtin, unused-variable
            NodesRepositoryInterface.delete.__doc__
        )
        row = self.client.select_one(f"SELECT count(url) FROM {TABLE_NAME}")
        if row is None:
            return 0

        return row[0]

    def get_urls(self) -> List[str]:
        __doc__ = (  # pylint: disable=redefined-builtin, unused-variable
            NodesRepositoryInterface.get_urls.__doc__
        )
        sql = f"SELECT url FROM {TABLE_NAME}"
        result_set = self.client.select(sql)
        list_ = []
        for row in result_set:
            list_.append(row[0])

        return list_


def get_fields_from_node(node: Node) -> dict:
    """
    Return a dict of supported fields with normalized value

    :param node: Node instance
    :return:
    """
    fields = {}
    for (key, value) in node.__dict__.items():
        if key.startswith("_"):
            continue
        elif isinstance(value, bool):
            value = 1 if value is True else 0
        fields[key] = value

    return fields


def get_node_from_row(row: tuple) -> Node:
    """
    Return a Node instance from a result set row

    :param row: Result set row
    :return:
    """
    values: List[Any] = []
    count = 0
    for value in row:
        if count == 7:
            values.append(value == 1)
        else:
            values.append(value)
        count += 1

    return Node(*values)



# Copyright 2021 Vincent Texier <vit@free.fr>
#
# This software is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
from gql import gql

from tikka.adapters.network.indexer.connection import IndexerConnection
from tikka.adapters.network.indexer.identities import IndexerIdentities
from tikka.adapters.network.indexer.transfers import IndexerTransfers
from tikka.domains.entities.indexer import Indexer
from tikka.interfaces.adapters.network.connection import NetworkConnectionError
from tikka.interfaces.adapters.network.indexer.indexer import (
    NetworkIndexerException,
    NetworkIndexerInterface,
)


class NetworkIndexer(NetworkIndexerInterface):
    """
    NetworkIndexer class
    """

    def __init__(self):
        """
        Init NetworkIndexer instance
        """
        self._connection = IndexerConnection()
        self._transfers = IndexerTransfers(self)
        self._identities = IndexerIdentities(self)

    @property
    def connection(self) -> IndexerConnection:
        __doc__ = (  # pylint: disable=redefined-builtin, unused-variable
            NetworkIndexerInterface.connection.__doc__
        )
        return self._connection

    def get(self) -> Indexer:
        __doc__ = (  # pylint: disable=redefined-builtin, unused-variable
            NetworkIndexerInterface.get.__doc__
        )
        if not self.connection.is_connected() or self.connection.client is None:
            raise NetworkIndexerException(NetworkConnectionError())

        query = gql(
            """
            query {
                block(
                    # distinctOn: [id]
                    limit: 1
                    offset: 0
                    orderBy: [{timestamp: DESC_NULLS_LAST }]
                    where: {}
                  ) {
                    height
                  }
            }
            """
        )
        try:
            result = self.connection.client.execute(query)
        except Exception as exception:
            raise NetworkIndexerException(exception)

        return Indexer(self.connection.url, block=result["block"][0]["height"])

    @property
    def identities(self) -> IndexerIdentities:
        """
        Return IndexerIdentities instance

        :return:
        """
        return self._identities

    @property
    def transfers(self) -> IndexerTransfers:
        """
        Return IndexerTransfers instance

        :return:
        """
        return self._transfers



# Copyright 2021 Vincent Texier <vit@free.fr>
#
# This software is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import abc
from typing import Any, Dict, List, Optional

from tikka.domains.entities.transfer import Transfer


class TransfersRepositoryInterface(abc.ABC):
    """
    TransfersRepositoryInterface class
    """

    COLUMN_ID = "transfer_id"
    COLUMN_ISSUER_ADDRESS = "transfer_issuer_address"
    COLUMN_ISSUER_IDENTITY_INDEX = "transfer_issuer_identity_index"
    COLUMN_ISSUER_IDENTITY_NAME = "transfer_issuer_identity_name"
    COLUMN_RECEIVER_ADDRESS = "transfer_receiver_address"
    COLUMN_RECEIVER_IDENTITY_INDEX = "transfer_receiver_identity_index"
    COLUMN_RECEIVER_IDENTITY_NAME = "transfer_receiver_identity_name"
    COLUMN_AMOUNT = "transfer_amount"
    COLUMN_TIMESTAMP = "transfer_timestamp"
    COLUMN_COMMENT = "transfer_comment"
    COLUMN_COMMENT_TYPE = "transfer_comment_type"

    SORT_ORDER_ASCENDING = "ASC"
    SORT_ORDER_DESCENDING = "DESC"

    @abc.abstractmethod
    def add(self, address: str, transfer: Transfer) -> None:
        """
        Add a new transfer to account history in repository

        :param address: Account address
        :param transfer: Transfer instance
        :return:
        """
        raise NotImplementedError

    @abc.abstractmethod
    def set_history(self, address: str, transfers: List[Transfer]) -> None:
        """
        Set account transfers history in repository

        :param address: Account address
        :param transfers: List of Transfer instances
        :return:
        """
        raise NotImplementedError

    @abc.abstractmethod
    def list(
        self,
        address: str,
        filters: Optional[Dict[str, Any]] = None,
        sort_column: str = COLUMN_TIMESTAMP,
        sort_order: str = SORT_ORDER_DESCENDING,
        limit: int = 10,
        offset: int = 0,
    ) -> List[Transfer]:
        """
        List transfers from and to address, from repository with optional filters and sort_column

        :param address: Account address
        :param filters: Dict with {column: value} filters or None
        :param sort_column: Sort column constant, default to COLUMN_TIMESTAMP
        :param sort_order: Sort order constant SORT_ORDER_ASCENDING or SORT_ORDER_DESCENDING
        :param limit: Number of rows to return
        :param offset: Offset of first row to return
        :return:
        """
        raise NotImplementedError

    @abc.abstractmethod
    def delete(self, transfer_id: str) -> None:
        """
        Delete transfer with transfer_id in repository

        :param transfer_id: Transfer ID to delete
        :return:
        """
        raise NotImplementedError

    @abc.abstractmethod
    def delete_all(self) -> None:
        """
        Delete all transfers in repository

        :return:
        """
        raise NotImplementedError

    @abc.abstractmethod
    def count(self, address) -> int:
        """
        Return total number of transfers issued and received by address

        :return:
        """
        raise NotImplementedError



# Copyright 2021 Vincent Texier <vit@free.fr>
#
# This software is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
from typing import List

from PyQt5.QtCore import (
    QAbstractItemModel,
    QAbstractListModel,
    QLocale,
    QModelIndex,
    QSize,
    Qt,
    QVariant,
)
from PyQt5.QtGui import QPainter, QPalette
from PyQt5.QtWidgets import QStyle, QStyledItemDelegate, QStyleOptionViewItem

from tikka.domains.application import Application
from tikka.domains.entities.constants import AMOUNT_UNIT_KEY
from tikka.domains.entities.transfer import Transfer
from tikka.interfaces.adapters.repository.transfers import TransfersRepositoryInterface
from tikka.slots.pyqt.entities.constants import SELECTED_UNIT_PREFERENCES_KEY
from tikka.slots.pyqt.widgets.account_transfer_row import AccountTransferRowWidget


class TransferItemDelegate(QStyledItemDelegate):
    """
    TransferItemDelegate class
    """

    def __init__(self):
        """
        Init TransferItemDelegate instance
        """
        super().__init__()

        # set item height >= AccountTransferRowWidget.height
        self.item_height = 71

    def sizeHint(self, option: QStyleOptionViewItem, index: QModelIndex):
        """
        Return item size

        Force height

        :param option: QStyleOptionViewItem instance
        :param index: QModelIndex instance
        :return:
        """
        return QSize(option.rect.width(), self.item_height)

    def paint(
        self, painter: QPainter, option: QStyleOptionViewItem, index: QModelIndex
    ) -> None:
        """
        Draw transfer row item in listview

        :param painter: QPainter instance
        :param option: QStyleOptionViewItem instance
        :param index: QModelIndex instance
        :return:
        """
        # Init style option
        style_option = QStyleOptionViewItem(option)
        self.initStyleOption(style_option, index)

        account_address_label = index.data(TransfersListModel.AccountAddressLabelRole)
        account_name_label = index.data(TransfersListModel.AccountNameRole)
        identity_name_label = index.data(TransfersListModel.IdentityNameRole)
        account_has_identity = index.data(TransfersListModel.AccountHasIdentityRole)
        datetime_label = index.data(TransfersListModel.DatetimeLabelRole)
        comment_type = index.data(TransfersListModel.CommentTypeRole)
        comment_label = index.data(TransfersListModel.CommentLabelRole)
        amount_label = index.data(TransfersListModel.AmountLabelRole)

        """ Replace item row by the widget """
        widget = AccountTransferRowWidget(
            account_address_label,
            account_name_label,
            identity_name_label,
            account_has_identity,
            datetime_label,
            amount_label,
            comment_label,
            comment_type,
        )  # Crée le widget

        # set widget background
        # widget.setAutoFillBackground(True)

        # Modify widget palette depending on behavior (hover, selection)
        palette = widget.palette()

        # 🔹 Check if the line number is even (alternate color effect)
        if index.row() % 2 == 1:  # Even lines use AlternateBase color
            base_color = option.palette.color(QPalette.AlternateBase)
        else:
            base_color = option.palette.color(QPalette.Base)

        if option.state & QStyle.State_Selected:
            palette.setColor(
                QPalette.Window, option.palette.highlight().color()
            )  # selection color
        elif option.state & QStyle.State_MouseOver:
            palette.setColor(
                QPalette.Window, option.palette.color(QPalette.Highlight).lighter(200)
            )  # Hover
        else:
            palette.setColor(
                QPalette.Window, base_color
            )  # normal or alternate background

        widget.setPalette(palette)  # Apply palette on widget
        widget.update()  # 🔥 Force display of new colors

        widget.resize(option.rect.size())  # Set size to the row size
        painter.save()
        painter.translate(option.rect.topLeft())  # Place the widget correctly
        widget.render(painter)
        painter.restore()


class TransfersListModel(QAbstractListModel):
    """
    TransfersListModel class that drives the population of list display
    """

    (
        AccountAddressLabelRole,
        AccountHasIdentityRole,
        AccountNameRole,
        IdentityNameRole,
        DatetimeLabelRole,
        CommentLabelRole,
        CommentTypeRole,
        AmountLabelRole,
    ) = range(Qt.UserRole, Qt.UserRole + 8)

    def __init__(self, application: Application, address: str, locale: QLocale):
        """
        Init TransfersListModel with account address to get account transfers list


        :param application: Application instance
        :param address: Account address
        :param locale: QLocale instance
        """
        super().__init__()

        self.application = application
        self._ = self.application.translator.gettext

        # get parent widget locale to localize balance display
        self.locale = locale

        # calculate balance display by unit preference
        unit_preference = self.application.repository.preferences.get(
            SELECTED_UNIT_PREFERENCES_KEY
        )
        if unit_preference is not None:
            self.amount = self.application.amounts.get_amount(unit_preference)
        else:
            self.amount = self.application.amounts.get_amount(AMOUNT_UNIT_KEY)

        self.address = address
        self.transfers: List[Transfer] = []

    def init_data(self, offset: int = 0, limit: int = 10):
        """
        Fill data from repository

        :param offset: Offset in DB list to paginate
        :param limit: Limit page size
        :return:
        """
        # calculate balance display by unit preference
        unit_preference = self.application.repository.preferences.get(
            SELECTED_UNIT_PREFERENCES_KEY
        )
        if unit_preference is not None:
            self.amount = self.application.amounts.get_amount(unit_preference)
        else:
            self.amount = self.application.amounts.get_amount(AMOUNT_UNIT_KEY)

        self.beginResetModel()
        self.transfers = self.application.transfers.repository.list(
            self.address,
            offset=offset,
            limit=limit,
            sort_column=TransfersRepositoryInterface.COLUMN_TIMESTAMP,
            sort_order=TransfersRepositoryInterface.SORT_ORDER_DESCENDING,
        )
        self.endResetModel()

    def rowCount(self, _: QModelIndex = QModelIndex()) -> int:
        """
        Return row count

        :param _: QModelIndex instance
        :return:
        """
        return len(self.transfers)

    def index(  # type: ignore
        self, row: int, column: int, parent: QModelIndex = QModelIndex()
    ) -> QModelIndex:
        """
        Return QModelIndex for row, column and parent

        :param row: Row index
        :param column: Column index
        :param parent: Parent QModelIndex instance
        :return:
        """
        if not QAbstractItemModel.hasIndex(self, row, column, parent):
            return QModelIndex()

        if row < len(self.transfers):
            return QAbstractItemModel.createIndex(
                self, row, column, self.transfers[row]
            )

        return QModelIndex()

    def data(self, index: QModelIndex, role: int = Qt.DisplayRole) -> QVariant:
        """
        Return data of cell for column index.column

        :param index: QModelIndex instance
        :param role: Item data role
        :return:
        """
        row = index.row()
        transfer = self.transfers[row]
        data = QVariant()

        if role == self.AccountAddressLabelRole:
            # return address issuer
            if transfer.issuer_address != self.address:
                return QVariant(
                    f"{transfer.issuer_address[:4]}…{transfer.issuer_address[-5:]}"
                )
            else:
                return QVariant(
                    f"{transfer.receiver_address[:4]}…{transfer.receiver_address[-5:]}"
                )

        if role == self.AccountHasIdentityRole:
            if transfer.issuer_address != self.address:
                return QVariant(transfer.issuer_identity_index is not None)
            else:
                return QVariant(transfer.receiver_identity_index is not None)

        if role == self.AccountNameRole:
            # return transfer received
            if transfer.issuer_address != self.address:
                account = self.application.accounts.get_by_address(
                    transfer.issuer_address
                )
                if account is not None and account.name is not None:
                    return QVariant(account.name)
                else:
                    return QVariant()

            # display transfer issued
            else:
                account = self.application.accounts.get_by_address(
                    transfer.receiver_address
                )
                if account is not None and account.name is not None:
                    return QVariant(account.name)
                else:
                    return QVariant()

        if role == self.IdentityNameRole:

            # return transfer received
            if transfer.issuer_address != self.address:
                if transfer.issuer_identity_index is not None:
                    return QVariant(
                        f"{transfer.issuer_identity_name}#{transfer.issuer_identity_index}"
                    )
                else:
                    return QVariant()

            # display transfer issued
            else:
                if transfer.receiver_identity_index is not None:
                    return QVariant(
                        f"{transfer.receiver_identity_name}#{transfer.receiver_identity_index}"
                    )
                else:
                    return QVariant()

        if role == self.DatetimeLabelRole:
            # datetime
            datetime_label = self.locale.toString(
                transfer.timestamp.astimezone(),
                QLocale.dateTimeFormat(self.locale, QLocale.ShortFormat),
            )
            return QVariant(datetime_label)

        if role == self.CommentTypeRole:
            return QVariant(transfer.comment_type)

        if role == self.CommentLabelRole:
            # comment
            comment_label = ""
            if transfer.comment is not None:
                if (
                    transfer.comment_type == "ASCII"
                    or transfer.comment_type == "UNICODE"
                ):
                    comment_label = transfer.comment
                elif transfer.comment_type == "RAW":
                    comment_label = (
                        f"RAW:{transfer.comment[:min(len(transfer.comment),10)]}..."
                        f"{transfer.comment[max(-len(transfer.comment),-10):]}"
                    )
                elif transfer.comment_type == "CID":
                    comment_label = f"CID:{transfer.comment}"

            return QVariant(comment_label)

        if role == self.AmountLabelRole:
            # display transfer received
            if transfer.issuer_address != self.address:
                amount_sign = ""
            else:
                amount_sign = "-"
            # amount
            amount = self.locale.toCurrencyString(
                self.amount.value(transfer.amount), self.amount.symbol()
            )
            return QVariant(f"{amount_sign}{amount}")

        return data



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'tikka/slots/pyqt/resources/gui/windows/account_create.ui'
#
# Created by: PyQt5 UI code generator 5.15.10
#
# WARNING: Any manual changes made to this file will be lost when pyuic5 is
# run again.  Do not edit this file unless you know what you are doing.


from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_AccountCreateDialog(object):
    def setupUi(self, AccountCreateDialog):
        AccountCreateDialog.setObjectName("AccountCreateDialog")
        AccountCreateDialog.resize(924, 497)
        self.verticalLayout = QtWidgets.QVBoxLayout(AccountCreateDialog)
        self.verticalLayout.setObjectName("verticalLayout")
        self.imagelabel = QtWidgets.QLabel(AccountCreateDialog)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.imagelabel.sizePolicy().hasHeightForWidth())
        self.imagelabel.setSizePolicy(sizePolicy)
        self.imagelabel.setMinimumSize(QtCore.QSize(0, 0))
        self.imagelabel.setMaximumSize(QtCore.QSize(200, 200))
        self.imagelabel.setSizeIncrement(QtCore.QSize(0, 0))
        self.imagelabel.setBaseSize(QtCore.QSize(0, 0))
        self.imagelabel.setText("")
        self.imagelabel.setPixmap(QtGui.QPixmap(":/images/safe_background"))
        self.imagelabel.setScaledContents(True)
        self.imagelabel.setAlignment(QtCore.Qt.AlignCenter)
        self.imagelabel.setObjectName("imagelabel")
        self.verticalLayout.addWidget(self.imagelabel, 0, QtCore.Qt.AlignHCenter)
        self.gridLayout = QtWidgets.QGridLayout()
        self.gridLayout.setObjectName("gridLayout")
        self.addressLabel = QtWidgets.QLabel(AccountCreateDialog)
        font = QtGui.QFont()
        font.setBold(True)
        font.setWeight(75)
        self.addressLabel.setFont(font)
        self.addressLabel.setAlignment(QtCore.Qt.AlignRight|QtCore.Qt.AlignTrailing|QtCore.Qt.AlignVCenter)
        self.addressLabel.setObjectName("addressLabel")
        self.gridLayout.addWidget(self.addressLabel, 5, 0, 1, 1)
        self.nameLabel = QtWidgets.QLabel(AccountCreateDialog)
        font = QtGui.QFont()
        font.setBold(True)
        font.setWeight(75)
        self.nameLabel.setFont(font)
        self.nameLabel.setAlignment(QtCore.Qt.AlignRight|QtCore.Qt.AlignTrailing|QtCore.Qt.AlignVCenter)
        self.nameLabel.setObjectName("nameLabel")
        self.gridLayout.addWidget(self.nameLabel, 6, 0, 1, 1)
        self.mnemonicLineEdit = QtWidgets.QLineEdit(AccountCreateDialog)
        self.mnemonicLineEdit.setReadOnly(True)
        self.mnemonicLineEdit.setObjectName("mnemonicLineEdit")
        self.gridLayout.addWidget(self.mnemonicLineEdit, 0, 1, 1, 1)
        self.mnemonicLabel = QtWidgets.QLabel(AccountCreateDialog)
        font = QtGui.QFont()
        font.setBold(True)
        font.setWeight(75)
        self.mnemonicLabel.setFont(font)
        self.mnemonicLabel.setAlignment(QtCore.Qt.AlignRight|QtCore.Qt.AlignTrailing|QtCore.Qt.AlignVCenter)
        self.mnemonicLabel.setObjectName("mnemonicLabel")
        self.gridLayout.addWidget(self.mnemonicLabel, 0, 0, 1, 1)
        self.nameLineEdit = QtWidgets.QLineEdit(AccountCreateDialog)
        self.nameLineEdit.setObjectName("nameLineEdit")
        self.gridLayout.addWidget(self.nameLineEdit, 6, 1, 1, 1)
        self.addressValueLabel = QtWidgets.QLabel(AccountCreateDialog)
        self.addressValueLabel.setText("")
        self.addressValueLabel.setTextInteractionFlags(QtCore.Qt.LinksAccessibleByMouse|QtCore.Qt.TextSelectableByKeyboard|QtCore.Qt.TextSelectableByMouse)
        self.addressValueLabel.setObjectName("addressValueLabel")
        self.gridLayout.addWidget(self.addressValueLabel, 5, 1, 1, 1)
        self.mnemonicLanguageComboBox = QtWidgets.QComboBox(AccountCreateDialog)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.mnemonicLanguageComboBox.sizePolicy().hasHeightForWidth())
        self.mnemonicLanguageComboBox.setSizePolicy(sizePolicy)
        self.mnemonicLanguageComboBox.setObjectName("mnemonicLanguageComboBox")
        self.gridLayout.addWidget(self.mnemonicLanguageComboBox, 2, 1, 1, 1)
        self.mnemonicLanguageLabel = QtWidgets.QLabel(AccountCreateDialog)
        font = QtGui.QFont()
        font.setBold(True)
        font.setItalic(False)
        font.setWeight(75)
        self.mnemonicLanguageLabel.setFont(font)
        self.mnemonicLanguageLabel.setAlignment(QtCore.Qt.AlignRight|QtCore.Qt.AlignTrailing|QtCore.Qt.AlignVCenter)
        self.mnemonicLanguageLabel.setObjectName("mnemonicLanguageLabel")
        self.gridLayout.addWidget(self.mnemonicLanguageLabel, 2, 0, 1, 1)
        self.verticalLayout.addLayout(self.gridLayout)
        self.changeButton = QtWidgets.QPushButton(AccountCreateDialog)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(7)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.changeButton.sizePolicy().hasHeightForWidth())
        self.changeButton.setSizePolicy(sizePolicy)
        icon = QtGui.QIcon()
        icon.addPixmap(QtGui.QPixmap(":/icons/refresh_icon"), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        self.changeButton.setIcon(icon)
        self.changeButton.setObjectName("changeButton")
        self.verticalLayout.addWidget(self.changeButton, 0, QtCore.Qt.AlignHCenter)
        self.horizontalLayout_2 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_2.setObjectName("horizontalLayout_2")
        self.saveMnemonicCheckBox = QtWidgets.QCheckBox(AccountCreateDialog)
        self.saveMnemonicCheckBox.setObjectName("saveMnemonicCheckBox")
        self.horizontalLayout_2.addWidget(self.saveMnemonicCheckBox)
        self.verticalLayout.addLayout(self.horizontalLayout_2)
        self.passwordFrame = QtWidgets.QFrame(AccountCreateDialog)
        self.passwordFrame.setFrameShape(QtWidgets.QFrame.NoFrame)
        self.passwordFrame.setFrameShadow(QtWidgets.QFrame.Raised)
        self.passwordFrame.setObjectName("passwordFrame")
        self.verticalLayout_2 = QtWidgets.QVBoxLayout(self.passwordFrame)
        self.verticalLayout_2.setObjectName("verticalLayout_2")
        self.horizontalLayout = QtWidgets.QHBoxLayout()
        self.horizontalLayout.setObjectName("horizontalLayout")
        spacerItem = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout.addItem(spacerItem)
        self.passwordLabel = QtWidgets.QLabel(self.passwordFrame)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.passwordLabel.sizePolicy().hasHeightForWidth())
        self.passwordLabel.setSizePolicy(sizePolicy)
        self.passwordLabel.setSizeIncrement(QtCore.QSize(0, 0))
        font = QtGui.QFont()
        font.setBold(True)
        font.setWeight(75)
        self.passwordLabel.setFont(font)
        self.passwordLabel.setAlignment(QtCore.Qt.AlignLeading|QtCore.Qt.AlignLeft|QtCore.Qt.AlignVCenter)
        self.passwordLabel.setObjectName("passwordLabel")
        self.horizontalLayout.addWidget(self.passwordLabel)
        self.passwordLineEdit = QtWidgets.QLineEdit(self.passwordFrame)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.passwordLineEdit.sizePolicy().hasHeightForWidth())
        self.passwordLineEdit.setSizePolicy(sizePolicy)
        self.passwordLineEdit.setMaxLength(20)
        self.passwordLineEdit.setAlignment(QtCore.Qt.AlignCenter)
        self.passwordLineEdit.setReadOnly(True)
        self.passwordLineEdit.setObjectName("passwordLineEdit")
        self.horizontalLayout.addWidget(self.passwordLineEdit)
        self.passwordChangeButton = QtWidgets.QPushButton(self.passwordFrame)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.passwordChangeButton.sizePolicy().hasHeightForWidth())
        self.passwordChangeButton.setSizePolicy(sizePolicy)
        self.passwordChangeButton.setIcon(icon)
        self.passwordChangeButton.setObjectName("passwordChangeButton")
        self.horizontalLayout.addWidget(self.passwordChangeButton)
        spacerItem1 = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout.addItem(spacerItem1)
        self.verticalLayout_2.addLayout(self.horizontalLayout)
        self.verticalLayout.addWidget(self.passwordFrame)
        self.buttonBox = QtWidgets.QDialogButtonBox(AccountCreateDialog)
        self.buttonBox.setOrientation(QtCore.Qt.Horizontal)
        self.buttonBox.setStandardButtons(QtWidgets.QDialogButtonBox.Cancel|QtWidgets.QDialogButtonBox.Ok)
        self.buttonBox.setObjectName("buttonBox")
        self.verticalLayout.addWidget(self.buttonBox, 0, QtCore.Qt.AlignHCenter)
        spacerItem2 = QtWidgets.QSpacerItem(20, 40, QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Expanding)
        self.verticalLayout.addItem(spacerItem2)

        self.retranslateUi(AccountCreateDialog)
        self.buttonBox.accepted.connect(AccountCreateDialog.accept) # type: ignore
        self.buttonBox.rejected.connect(AccountCreateDialog.reject) # type: ignore
        QtCore.QMetaObject.connectSlotsByName(AccountCreateDialog)
        AccountCreateDialog.setTabOrder(self.mnemonicLineEdit, self.nameLineEdit)
        AccountCreateDialog.setTabOrder(self.nameLineEdit, self.changeButton)
        AccountCreateDialog.setTabOrder(self.changeButton, self.saveMnemonicCheckBox)
        AccountCreateDialog.setTabOrder(self.saveMnemonicCheckBox, self.passwordLineEdit)
        AccountCreateDialog.setTabOrder(self.passwordLineEdit, self.passwordChangeButton)

    def retranslateUi(self, AccountCreateDialog):
        #  _translate = QtCore.QCoreApplication.translate
        AccountCreateDialog.setWindowTitle(_("Create account"))
        self.addressLabel.setText(_("Address"))
        self.nameLabel.setText(_("Name"))
        self.mnemonicLabel.setText(_("Mnemonic"))
        self.mnemonicLanguageLabel.setText(_("Mnemonic language"))
        self.changeButton.setText(_("Change"))
        self.saveMnemonicCheckBox.setText(_("I have saved my mnemonic phrase in a safety place"))
        self.passwordLabel.setText(_("Password"))
        self.passwordChangeButton.setText(_("Change"))
import tikka.slots.pyqt.resources.icons.index_rc



# Copyright 2021 Vincent Texier <vit@free.fr>
#
# This software is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from tikka.domains.entities.currency import Currency
from tikka.interfaces.adapters.network.node.currency import NodeCurrencyInterface
from tikka.interfaces.adapters.repository.currencies import (
    CurrenciesRepositoryInterface,
)
from tikka.interfaces.adapters.repository.currency import CurrencyRepositoryInterface


class Currencies:
    """
    Currencies domain class
    """

    def __init__(
        self,
        repository: CurrenciesRepositoryInterface,
        currency_repository: CurrencyRepositoryInterface,
        network: NodeCurrencyInterface,
        code_name: str,
    ):
        """
        Init Currencies instance

        :param repository: CurrenciesRepositoryInterface instance
        :param currency_repository: CurrencyRepositoryInterface instance
        :param network: NodeCurrencyInterface instance
        :param code_name: Code name of current currency instance
        """
        self.repository = repository
        self.currency_repository = currency_repository
        self.network = network
        self._current = self._get(code_name)

    def _get(self, code_name: str) -> Currency:
        """
        Get Currency instance from code_name

        :param code_name: Currency code name
        :return:
        """
        currency = self.currency_repository.get(code_name)
        # if first connection on this currency...
        if currency is None:
            # get currency from currencies yaml file
            currency = self.repository.get(code_name)
            if currency is None:
                raise ValueError("code name is unknown")
            self.currency_repository.add(currency)

        return currency

    def get_current(self) -> Currency:
        """
        Return current Currency instance

        :return:
        """
        return self._current

    def set_current(self, code_name: str) -> None:
        """
        Set current Currency instance by code name

        :param code_name: Currency code name
        :return:
        """
        self._current = self._get(code_name)

    def code_names(self) -> list:
        """
        Return list of currency code names

        :return:
        """
        return self.repository.code_names()

    def names(self) -> list:
        """
        Return list of currency names

        :return:
        """
        return self.repository.names()

    def get_entry_point_urls(
        self,
    ) -> dict:
        """
        Return current currency entry point urls

        :return:
        """
        return self.repository.get_entry_point_urls(self.get_current().code_name)

    def network_update_properties(self):
        """
        Get currency properties from network and update currency

        :return:
        """
        currency = self.network.get_instance()
        self.currency_repository.update(currency)



# Copyright 2021 Vincent Texier <vit@free.fr>
#
# This software is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from typing import Callable, Optional

from PyQt5.QtCore import QMutex, QObject, QThread, pyqtSignal


class QWorker(QObject):
    finished = pyqtSignal()

    def __init__(self, call: Callable, mutex: QMutex) -> None:
        """
        Init QWorker instance with function to call and mutex (thread lock handling)

        :param call: Function to call
        :param mutex: QMutex instance to handle thread locks
        """
        super().__init__()

        self.call = call
        self.mutex = mutex

    def run(self) -> None:
        """
        Run call function

        :return:
        """
        self.mutex.lock()
        self.call()
        self.mutex.unlock()

        self.finished.emit()


class AsyncQWorker(QThread):
    def __init__(
        self, call: Callable, mutex: QMutex, parent: Optional[QObject] = None
    ) -> None:
        """
        Init QWorker instance with function to call

        :param call: Function to call
        :param mutex: QMutex instance to handle thread locks
        :param parent: Parent QObject instance
        """
        super().__init__(parent)
        self.worker = QWorker(call, mutex)
        self.worker.moveToThread(self)
        # run worker when thread is started
        self.started.connect(self.worker.run)
        # quit thread when worker is finished
        self.worker.finished.connect(self.quit)



# Copyright 2021 Vincent Texier <vit@free.fr>
#
# This software is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
from datetime import datetime
from typing import List, Optional

from dateutil import parser
from gql import gql

from tikka.domains.entities.transfer import Transfer
from tikka.interfaces.adapters.network.connection import NetworkConnectionError
from tikka.interfaces.adapters.network.indexer.indexer import NetworkIndexerException
from tikka.interfaces.adapters.network.indexer.transfers import (
    IndexerTransfersException,
    IndexerTransfersInterface,
)


class IndexerTransfers(IndexerTransfersInterface):
    """
    IndexerTransfers class
    """

    def list(
        self,
        address: str,
        limit: int,
        offset: int = 0,
        sort_column: str = IndexerTransfersInterface.COLUMN_TIMESTAMP,
        sort_order: str = IndexerTransfersInterface.SORT_ORDER_DESCENDING,
        from_datetime: Optional[datetime] = None,
        to_datetime: Optional[datetime] = None,
    ) -> List[Transfer]:
        __doc__ = (  # pylint: disable=redefined-builtin, unused-variable
            IndexerTransfersInterface.list.__doc__
        )

        if (
            not self.indexer.connection.is_connected()
            or self.indexer.connection.client is None
        ):
            raise IndexerTransfersException(NetworkConnectionError())

        # Construire la condition de filtrage pour la date
        date_filter = []
        if from_datetime:
            date_filter.append(
                f'{{ timestamp: {{ _gte: "{from_datetime.isoformat()}" }} }}'
            )
        if to_datetime:
            date_filter.append(
                f'{{ timestamp: {{ _lte: "{to_datetime.isoformat()}" }} }}'
            )

        date_filter_str = (
            "_and: [" + ", ".join(date_filter) + "]" if date_filter else ""
        )

        # Construire la requête GraphQL
        query = gql(
            f"""
            query {{
              transfer(
                limit: {limit}
                offset: {offset}
                orderBy: {{
                    {sort_column}: {sort_order}
                }}
                where: {{
                    _or: [
                        {{ fromId: {{ _eq: \"{address}\" }} }}
                        {{ toId: {{ _eq: \"{address}\" }} }}
                    ]
                    {date_filter_str}
                }}
              ) {{
                    id
                    fromId
                    from {{
                      identity {{
                        index
                        name
                      }}
                    }}
                    toId
                    to {{
                      identity {{
                        index
                        name
                      }}
                    }}
                    amount
                    timestamp
                    comment {{
                      type
                      remark
                      remarkBytes
                    }}
                }}
            }}
        """
        )

        try:
            result = self.indexer.connection.client.execute(query)
        except Exception as exception:
            raise NetworkIndexerException(exception)

        transfers = []
        for transfer in result["transfer"]:
            issuer_identity_index = (
                transfer["from"]["identity"]["index"]
                if transfer["from"]["identity"]
                else None
            )
            issuer_identity_name = (
                transfer["from"]["identity"]["name"]
                if transfer["from"]["identity"]
                else None
            )
            receiver_identity_index = (
                transfer["to"]["identity"]["index"]
                if transfer["to"]["identity"]
                else None
            )
            receiver_identity_name = (
                transfer["to"]["identity"]["name"]
                if transfer["to"]["identity"]
                else None
            )

            comment = None
            comment_type = None
            if transfer["comment"]:
                comment = (
                    transfer["comment"]["remarkBytes"]
                    if transfer["comment"]["type"] == "RAW"
                    else transfer["comment"]["remark"]
                )
                comment_type = transfer["comment"]["type"]

            transfers.append(
                Transfer(
                    id=transfer["id"],
                    issuer_address=transfer["fromId"],
                    issuer_identity_index=issuer_identity_index,
                    issuer_identity_name=issuer_identity_name,
                    receiver_address=transfer["toId"],
                    receiver_identity_index=receiver_identity_index,
                    receiver_identity_name=receiver_identity_name,
                    amount=transfer["amount"],
                    timestamp=parser.parse(transfer["timestamp"]),
                    comment=comment,
                    comment_type=comment_type,
                )
            )

        return transfers

    def count(self, address) -> int:
        __doc__ = (  # pylint: disable=redefined-builtin, unused-variable
            IndexerTransfersInterface.count.__doc__
        )
        if (
            not self.indexer.connection.is_connected()
            or self.indexer.connection.client is None
        ):
            raise IndexerTransfersException(NetworkConnectionError())

        query = gql(
            """
            query {
                transferAggregate (
                    where: {
                        _or: [
                        {fromId: {_eq: \""""
            + address
            + """\"}}
                        {toId: {_eq: \""""
            + address
            + """\"}}
                        ]
                    }
                ) {
                    aggregate {
                        count
                    }
                }
            }
            """
        )
        try:
            result = self.indexer.connection.client.execute(query)
        except Exception as exception:
            raise NetworkIndexerException(exception)

        # {
        #     "transferAggregate": {
        #       "aggregate": {
        #         "count": 3
        #       }
        #     }
        # }
        return result["transferAggregate"]["aggregate"]["count"]



# Copyright 2021 Vincent Texier <vit@free.fr>
#
# This software is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
from dataclasses import dataclass
from pathlib import Path
from typing import Optional

from tikka.libs.signing_key_v1 import SigningKey

WALLET_TYPES = {
    "PUBSEC": "PubSec: a file with your secret and public key",
    "WIF": "WIF: a file with your seed",
    "EWIF": "EWIF: an encrypted file with your seed, need a password",
    "DEWIF": "DEWIF: an encrypted file with your seed, need a password",
}


@dataclass
class V1FileWallet:
    path: Path
    type: str
    is_encrypted: bool = False
    signing_key: Optional[SigningKey] = None


@dataclass
class Wallet:
    address: str
    crypto_type: int
    encrypted_private_key: str
    encryption_nonce: str
    encryption_mac_tag: str



# Copyright 2021 Vincent Texier <vit@free.fr>
#
# This software is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
import logging
import sys
from typing import Optional

from PyQt5.QtCore import QModelIndex, QMutex, QPoint, QRect, QSize, Qt
from PyQt5.QtGui import QBrush, QColor, QFontMetrics, QPainter, QPixmap
from PyQt5.QtWidgets import (
    QApplication,
    QMainWindow,
    QStyledItemDelegate,
    QStyleOptionViewItem,
    QWidget,
)

from tikka.domains.application import Application
from tikka.domains.entities.constants import AMOUNT_UNIT_KEY, DATA_PATH
from tikka.domains.entities.events import AccountEvent, CurrencyEvent, UnitEvent
from tikka.interfaces.adapters.repository.accounts import AccountsRepositoryInterface
from tikka.slots.pyqt.entities.constants import (
    ACCOUNTS_TABLE_CATEGORY_FILTER_PREFERENCES_KEY,
    ACCOUNTS_TABLE_SORT_COLUMN_PREFERENCES_KEY,
    ACCOUNTS_TABLE_SORT_ORDER_PREFERENCES_KEY,
    ACCOUNTS_TABLE_WALLET_FILTER_PREFERENCES_KEY,
    NUMERIC_DISPLAY_COLOR_BLUE,
    SELECTED_UNIT_PREFERENCES_KEY,
)
from tikka.slots.pyqt.entities.worker import AsyncQWorker
from tikka.slots.pyqt.models.account_table import AccountTableModel, BalanceStyledRole
from tikka.slots.pyqt.resources.gui.widgets.account_table_rc import (
    Ui_AccountTableWidget,
)
from tikka.slots.pyqt.widgets.account_menu import AccountPopupMenu


class IconDelegate(QStyledItemDelegate):
    """
    IconDelegate class to center icons in table view
    """

    def initStyleOption(self, option: QStyleOptionViewItem, index: QModelIndex):
        """
        Init Style

        :param option: QStyleOptionViewItem instance
        :param index: QModelIndex instance
        :return:
        """
        super().initStyleOption(option, index)
        # resize width to actual width of icon
        option.decorationSize.setWidth(option.rect.width())


class IconTextDelegate(QStyledItemDelegate):
    def __init__(self, parent=None):
        super().__init__(parent)

    def paint(self, painter: QPainter, option, index):
        """
        Draw a QLabel in the cell


        :param painter: QPainter instance
        :param option:
        :param index:
        :return:
        """
        if not index.isValid():
            return

        # Récupérer l'icône et le texte
        icon = index.data(Qt.DecorationRole)  # Récupère l'icône
        text: str = index.data(Qt.DisplayRole)  # Récupère le texte

        ###########################
        # PURE PAINT
        painter.save()

        # Dessiner l'icône
        if isinstance(icon, QPixmap) and not icon.isNull():
            painter.setPen(QColor("blue"))
            icon_rect = option.rect.adjusted(5, 5, -5, -5)  # Espace autour
            icon_pixmap = icon.scaled(
                16, 16, Qt.KeepAspectRatio
            )  # Ajuste la taille de l'icône
            painter.drawPixmap(icon_rect.left(), icon_rect.top(), icon_pixmap)

            # Décaler le texte pour ne pas recouvrir l'icône
            text_rect: QRect = option.rect.adjusted(25, 0, 0, 0)
        else:
            painter.setPen(QColor("red"))
            text_rect = option.rect

        # set member colors
        painter.drawText(text_rect, int(Qt.AlignVCenter | Qt.AlignLeft), text)

        painter.restore()

        ###########################
        # QLABEL DISPLAY DO NOT WORK !!!
        # # Créer un QLabel temporaire pour afficher l'icône + le texte
        # label = QLabel(text)
        # if isinstance(icon, QPixmap):
        #     label.setPixmap(icon)
        #
        # # label.resize(option.rect.size())  # Set size to the row size
        # label.setAlignment(Qt.AlignLeft | Qt.AlignVCenter)
        # # # Appliquer le style et dessiner le QLabel
        # # #label.setStyleSheet("padding: 2px;")  # Ajoute un petit padding
        # # label.setAlignment(Qt.AlignLeft | Qt.AlignVCenter)
        # label.resize(option.rect.size())  # Ajuste la taille au cadre de la cellule
        # # label.render(painter, option.rect.topLeft())
        #
        # painter.save()
        # painter.translate(option.rect.topLeft())  # Place the widget correctly
        # label.render(painter)
        # painter.restore()

    def sizeHint(self, option, index):
        """
        Retourne la taille suggérée pour la cellule en fonction du texte et de l'icône.
        """
        if not index.isValid():
            return QSize()

        text = index.data(Qt.DisplayRole)  # Texte de la cellule
        icon = index.data(Qt.DecorationRole)  # Icône de la cellule

        fm = option.fontMetrics  # Récupère les métriques de police
        text_width = fm.horizontalAdvance(text) + 10  # Largeur du texte + padding
        text_height = fm.height()  # Hauteur du texte

        icon_size = QSize(0, 0)
        if isinstance(icon, QPixmap) and not icon.isNull():
            icon_size = icon.size()  # Taille réelle de l'icône

        # Largeur totale = icône + texte + espacement
        total_width = (
            icon_size.width() + text_width + 10
            if not icon_size.isEmpty()
            else text_width
        )
        total_height = max(text_height, icon_size.height()) + 6  # Ajuste la hauteur

        return QSize(total_width, total_height)


class StyledBalanceItemDelegate(QStyledItemDelegate):
    """
    Class used to display balance with a custom style
    """

    def paint(
        self, painter: QPainter, option: QStyleOptionViewItem, index: QModelIndex
    ) -> None:
        """
        Return balance with a custom style

        :param painter: QPainter instance
        :param option: QStyleOptionViewItem instance
        :param index: QModelIndex instance
        :return:
        """
        if not index.isValid():
            return
        super().paint(painter, option, index)

        # draw balance with style
        balance_styled: str = index.data(BalanceStyledRole)
        if balance_styled:
            painter.save()

            # Enable anti-aliasing for smoother rendering
            painter.setRenderHint(QPainter.Antialiasing)

            # Set text color and font
            painter.setPen(Qt.white)
            painter.setFont(option.font)

            # Get text size using boundingRect
            font_metrics = QFontMetrics(option.font)
            text_rect = font_metrics.boundingRect(balance_styled)

            # Define padding and right margin
            padding_x, padding_y = 8, 4
            right_margin = 10  # Adjust this value for more space on the right

            # Adjust background rectangle size to include the right margin
            bg_width = text_rect.width() + 2 * padding_x + right_margin
            bg_height = text_rect.height() + 2 * padding_y

            # Position the rectangle to be right-aligned
            rect_x = option.rect.right() - bg_width - 5  # 5px margin from the right
            rect_y = option.rect.center().y() - bg_height // 2  # Centered vertically

            # Draw rounded rectangle
            painter.setBrush(
                QBrush(QColor(NUMERIC_DISPLAY_COLOR_BLUE))
            )  # Blue background
            # painter.setPen(Qt.NoPen)
            painter.drawRoundedRect(QRect(rect_x, rect_y, bg_width, bg_height), 10, 10)

            # Draw text inside the rectangle (horizontally centered with right margin)
            painter.drawText(
                QRect(
                    rect_x + padding_x,
                    rect_y + padding_y,
                    bg_width - padding_x - right_margin,
                    bg_height - 2 * padding_y,
                ),
                Qt.AlignCenter,
                balance_styled,
            )

            painter.restore()

    def sizeHint(self, option, index):
        """
        Return the appropriate size for the cell based on the content.
        """
        balance_styled: str = index.data(BalanceStyledRole) or ""
        font_metrics = QFontMetrics(option.font)

        # Calculate text size
        text_rect = font_metrics.boundingRect(balance_styled)

        # Define padding and right margin (same as used in paint)
        padding_x, padding_y = 8, 4
        right_margin = 10
        width = text_rect.width() + 2 * padding_x + right_margin
        height = text_rect.height() + 2 * padding_y

        return QSize(width, height)


class AccountTableWidget(QWidget, Ui_AccountTableWidget):
    """
    AccountTableWidget class
    """

    def __init__(
        self,
        application: Application,
        mutex: QMutex,
        parent: Optional[QWidget] = None,
    ) -> None:
        """
        Init AccountTableWidget instance

        :param application: Application instance
        :param mutex: QMutex instance
        :param parent: QWidget instance
        """
        super().__init__(parent=parent)
        self.setupUi(self)

        self.application = application
        self.mutex = mutex
        self.account_table_model = AccountTableModel(self.application, self.locale())
        self._ = self.application.translator.gettext

        # use icon delegate to center icons
        wallet_icon_delegate = IconDelegate(self.tableView)
        self.tableView.setItemDelegateForColumn(0, wallet_icon_delegate)
        self.identity_icon_text_delegate = IconTextDelegate()
        self.tableView.setItemDelegateForColumn(1, self.identity_icon_text_delegate)
        self.styled_balance_item_delegate = StyledBalanceItemDelegate()
        self.tableView.setItemDelegateForColumn(3, self.styled_balance_item_delegate)

        # setup table view
        self.tableView.setModel(self.account_table_model)
        # set header sort column and sort order icon
        sort_column_preference = self.application.repository.preferences.get(
            ACCOUNTS_TABLE_SORT_COLUMN_PREFERENCES_KEY
        )
        if sort_column_preference is not None:
            sort_column_index = int(sort_column_preference)
        else:
            sort_column_index = -1
        repository_sort_order = self.application.repository.preferences.get(
            ACCOUNTS_TABLE_SORT_ORDER_PREFERENCES_KEY
        )
        self.tableView.horizontalHeader().setSortIndicator(
            sort_column_index,
            Qt.AscendingOrder
            if repository_sort_order == AccountsRepositoryInterface.SORT_ORDER_ASCENDING
            else Qt.DescendingOrder,
        )
        self.tableView.setSortingEnabled(True)
        self.init_category_filter()
        self.init_wallet_filter()
        self.tableView.resizeColumnsToContents()
        self.tableView.resizeRowsToContents()

        ##############################
        # ASYNC METHODS
        ##############################
        # Create a QWorker object
        self.fetch_from_network_async_qworker = AsyncQWorker(
            self.fetch_from_network, self.mutex
        )
        self.fetch_from_network_async_qworker.finished.connect(
            self._on_finished_fetch_from_network
        )

        # events
        self.refreshButton.clicked.connect(self.fetch_from_network_async_qworker.start)

        # table events
        self.tableView.customContextMenuRequested.connect(self.on_context_menu)
        self.tableView.model().dataChanged.connect(self.on_data_changed)
        self.categoryComboBox.currentIndexChanged.connect(
            self.on_category_filter_changed
        )
        self.walletComboBox.currentIndexChanged.connect(self.on_wallet_filter_changed)

        # application events
        self.application.event_dispatcher.add_event_listener(
            UnitEvent.EVENT_TYPE_CHANGED, self.on_unit_event
        )
        self.application.event_dispatcher.add_event_listener(
            CurrencyEvent.EVENT_TYPE_CHANGED, self.on_currency_event
        )
        self.application.event_dispatcher.add_event_listener(
            AccountEvent.EVENT_TYPE_ADD, self.on_add_account_event
        )
        self.application.event_dispatcher.add_event_listener(
            AccountEvent.EVENT_TYPE_DELETE, self.on_delete_account_event
        )
        self.application.event_dispatcher.add_event_listener(
            AccountEvent.EVENT_TYPE_UPDATE, self.on_update_account_event
        )

        self._update_ui()

    def init_category_filter(self):
        """
        Fill category filter combo box with all categories and a blank choice

        :return:
        """
        # init combo box
        self.categoryComboBox.addItem(self._("No filter"), userData=None)
        for category in self.application.categories.list_all():
            self.categoryComboBox.addItem(category.name, userData=category.id.hex)

        # set current category filter from preferences
        category_id = self.application.repository.preferences.get(
            ACCOUNTS_TABLE_CATEGORY_FILTER_PREFERENCES_KEY
        )
        self.categoryComboBox.setCurrentIndex(
            self.categoryComboBox.findData(category_id)
        )

    def on_category_filter_changed(self, _):
        """
        Update preferences with category filter for table view

        :return:
        """
        category_id = self.categoryComboBox.currentData()
        self.application.repository.preferences.set(
            ACCOUNTS_TABLE_CATEGORY_FILTER_PREFERENCES_KEY,
            None if category_id is None else str(category_id),
        )
        self._update_model()
        self._update_ui()

    def init_wallet_filter(self):
        """
        Fill wallet filter combo box with None (with and without), True (with) or False (without)

        :return:
        """
        # init combo box
        self.walletComboBox.addItem(self._("No filter"), userData=None)
        self.walletComboBox.addItem(self._("Yes"), userData=True)
        self.walletComboBox.addItem(self._("No"), userData=False)

        # set current category filter from preferences
        wallet_filter_preference = self.application.repository.preferences.get(
            ACCOUNTS_TABLE_WALLET_FILTER_PREFERENCES_KEY
        )
        wallet_filter = None
        if wallet_filter_preference is not None:
            # preference store boolean as integer in a string column ("0" or "1")
            # convert it to boolean
            wallet_filter = int(wallet_filter_preference) == 1
        self.walletComboBox.setCurrentIndex(self.walletComboBox.findData(wallet_filter))

    def on_wallet_filter_changed(self, _):
        """
        Update preferences with wallet filter for table view

        :return:
        """
        wallet_filter = self.walletComboBox.currentData()
        self.application.repository.preferences.set(
            ACCOUNTS_TABLE_WALLET_FILTER_PREFERENCES_KEY, wallet_filter
        )
        self._update_model()
        self._update_ui()

    def on_data_changed(self, _):
        """
        Triggered when data changed in table model

        :return:
        """
        self._update_model()

    def _update_model(self):
        """
        Update all data in model

        :return:
        """
        self.tableView.model().init_data()
        self.tableView.resizeColumnsToContents()
        self.tableView.resizeRowsToContents()

    def _update_ui(self):
        """
        Update GUI

        :return:
        """
        table_total_balance = sum(
            [
                table_view_row.balance
                for table_view_row in self.account_table_model.table_view_data
                if table_view_row.balance is not None
            ]
        )

        unit_preference = self.application.repository.preferences.get(
            SELECTED_UNIT_PREFERENCES_KEY
        )
        if unit_preference is not None:
            amount = self.application.amounts.get_amount(unit_preference)
        else:
            amount = self.application.amounts.get_amount(AMOUNT_UNIT_KEY)

        self.totalBalanceValueLabel.setText(
            self.locale().toCurrencyString(
                amount.value(table_total_balance),
                amount.symbol(),
            )
        )

    def on_unit_event(self, _):
        """
        When a unit event is triggered

        :return:
        """
        self._update_model()
        self._update_ui()

    def on_currency_event(self, _):
        """
        When a currency event is triggered

        :return:
        """
        self._update_model()
        self._update_ui()

    def on_add_account_event(self, _):
        """
        Add account row when account is created

        :return:
        """
        self._update_model()
        self._update_ui()

    def on_delete_account_event(self, _):
        """
        Remove account row when account is deleted

        :return:
        """
        self._update_model()
        self._update_ui()

    def on_update_account_event(self, _):
        """
        Update account row when account is updated

        :return:
        """
        self._update_model()
        self._update_ui()

    def on_context_menu(self, position: QPoint):
        """
        When right button on table widget

        :param position: QPoint instance
        :return:
        """
        # get selected account
        table_view_row = self.tableView.currentIndex().internalPointer()
        account = self.application.accounts.get_by_address(table_view_row.address)
        if account is not None:
            # display popup menu at click position
            AccountPopupMenu(self.application, account, self.mutex, self).exec_(
                self.tableView.mapToGlobal(position)
            )

    def fetch_from_network(self):
        """
        Fetch table model accounts data from the network

        :return:
        """
        self.refreshButton.setEnabled(False)
        self.errorLabel.setText("")
        accounts = self.application.accounts.get_list()
        try:
            self.application.accounts.network_update_balances(accounts)
        except Exception as exception:
            self.errorLabel.setText(self._(str(exception)))

        addresses = [account.address for account in accounts]
        try:
            self.application.identities.network_update_identities(addresses)
        except Exception as exception:
            self.errorLabel.setText(self._(str(exception)))
        else:
            identity_indice = self.application.identities.list_indice()
            try:
                self.application.smiths.network_update_smiths(identity_indice)
            except Exception as exception:
                self.errorLabel.setText(self._(str(exception)))

    def _on_finished_fetch_from_network(self):
        """
        Triggered when async request fetch_from_network is finished

        :return:
        """
        logging.debug("Account table widget update")

        self.refreshButton.setEnabled(True)
        self._update_model()
        self._update_ui()


if __name__ == "__main__":
    qapp = QApplication(sys.argv)

    application_ = Application(DATA_PATH)

    main_window = QMainWindow()
    main_window.show()

    main_window.setCentralWidget(
        AccountTableWidget(application_, QMutex(), main_window)
    )

    sys.exit(qapp.exec_())



# Copyright 2021 Vincent Texier <vit@free.fr>
#
# This software is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
import abc
from typing import Any


class ConfigInterface(abc.ABC):
    """
    ConfigInterface class
    """

    CURRENCY_KEY = "currency"
    LANGUAGE_KEY = "language"
    RANDOM_CONNECTION_AT_START_KEY = "random_connection_at_start"

    @abc.abstractmethod
    def set(self, name: str, value: Any) -> Any:
        """
        Set named parameter to value

        :param name: Name of parameter
        :param value: New value
        :return:
        """
        raise NotImplementedError

    @abc.abstractmethod
    def get(self, name: str) -> Any:
        """
        Get value of named parameter

        :param name: Name of parameter
        :return:
        """
        raise NotImplementedError



# Copyright 2021 Vincent Texier <vit@free.fr>
#
# This software is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
from functools import reduce
from typing import Any, Dict, List, Optional
from uuid import UUID

from dateutil import parser
from sql import Column, Flavor, Table
from sql.conditionals import Case
from sql.operators import And, Equal, NotEqual

from tikka.adapters.repository.categories import TABLE_NAME as CATEGORIES_TABLE_NAME
from tikka.adapters.repository.identities import TABLE_NAME as IDENTITIES_TABLE_NAME
from tikka.adapters.repository.wallets import TABLE_NAME as WALLETS_TABLE_NAME
from tikka.domains.entities.account import Account
from tikka.interfaces.adapters.repository.accounts import AccountsRepositoryInterface
from tikka.interfaces.adapters.repository.db_repository import DBRepositoryInterface

TABLE_NAME = "accounts"

# create sql table wrapper
sql_accounts_table = Table(TABLE_NAME)

CATEGORIES_TABLE_NB_COLUMNS = 4
CATEGORY_NAME_ALIAS_COLUMN = "category_name_alias"
CATEGORY_ID_ALIAS_COLUMN = "category_id_alias"
IDENTITY_NAME_ALIAS_COLUMN = "identity_name_alias"


class DBAccountsRepository(AccountsRepositoryInterface, DBRepositoryInterface):
    """
    DBAccountsRepository class
    """

    def list(
        self,
        filters: Optional[Dict[str, Any]] = None,
        sort_column: Optional[str] = None,
        sort_order: str = AccountsRepositoryInterface.SORT_ORDER_ASCENDING,
    ) -> List[Account]:
        __doc__ = (  # pylint: disable=redefined-builtin, unused-variable
            AccountsRepositoryInterface.list.__doc__
        )

        sql_columns = {
            AccountsRepositoryInterface.COLUMN_ADDRESS: sql_accounts_table.address,
            AccountsRepositoryInterface.COLUMN_NAME: sql_accounts_table.name,
            AccountsRepositoryInterface.COLUMN_CRYPTO_TYPE: sql_accounts_table.crypto_type,
            AccountsRepositoryInterface.COLUMN_BALANCE: sql_accounts_table.balance,
            AccountsRepositoryInterface.COLUMN_PATH: sql_accounts_table.path,
            AccountsRepositoryInterface.COLUMN_ROOT: sql_accounts_table.root,
            AccountsRepositoryInterface.COLUMN_FILE_IMPORT: sql_accounts_table.file_import,
            AccountsRepositoryInterface.COLUMN_CATEGORY_ID: sql_accounts_table.category_id,
            AccountsRepositoryInterface.COLUMN_TOTAL_TRANSFERS_COUNT: sql_accounts_table.total_transfers_count,
        }

        # if sort column...
        if sort_column is not None:
            # set sort column
            sql_sort_colum: Column = sql_columns[sort_column]
            # create select query wrapper with order by
            sql_select = sql_accounts_table.select(
                order_by=sql_sort_colum.asc
                if sort_order == AccountsRepositoryInterface.SORT_ORDER_ASCENDING
                else sql_sort_colum.desc,
            )
        else:
            #  create select query wrapper without order by
            sql_select = sql_accounts_table.select()

        # create where conditions
        conditions = []
        if filters is not None:
            for key, value in filters.items():
                if key == AccountsRepositoryInterface.COLUMN_ROOT:
                    if value is True:
                        conditions.append(Equal(sql_accounts_table.root, None))
                    elif value is False:
                        conditions.append(NotEqual(sql_accounts_table.root, None))
                    elif isinstance(value, str):
                        conditions.append(
                            sql_accounts_table.root
                            == filters[AccountsRepositoryInterface.COLUMN_ROOT]
                        )
                else:
                    conditions.append(sql_columns[key] == value)

        # conditions are added with and operator
        def and_(a, b):
            return And((a, b))

        if len(conditions) > 0:
            sql_select.where = reduce(and_, conditions)  # type: ignore

        # config sql with ? as param style
        Flavor.set(Flavor(paramstyle="qmark"))

        sql, args = tuple(sql_select)
        result_set = self.client.select(sql, args)

        list_ = []
        for row in result_set:
            list_.append(get_account_from_row(row))

        return list_

    def table_view(
        self,
        filters: Optional[Dict[str, Any]] = None,
        sort_column_index: Optional[int] = None,
        sort_order: str = AccountsRepositoryInterface.SORT_ORDER_ASCENDING,
    ) -> List[AccountsRepositoryInterface.TableViewRow]:
        __doc__ = (  # pylint: disable=redefined-builtin, unused-variable
            AccountsRepositoryInterface.table_view.__doc__
        )
        # we need to get the category of root account for derived account
        # in order to do that, we join :
        # - the category table for root accounts
        # - the account table to get root account of derived accounts
        # - the category table for the joined account table to category of the root account of derived accounts

        # we also need to sort/filter by wallet column (stored or not)
        # in order to do that, we join :
        # - the wallet table

        # SELECT "e"."address", "a"."identity_index", "a"."balance", "a"."name", "a"."address", "a"."path", "a"."root",
        # "a"."crypto_type",
        # CASE WHEN ("a"."root" IS NULL) THEN "b"."name" ELSE "d"."name" END AS "category_name_alias",
        # CASE WHEN ("a"."root" IS NULL) THEN "b"."id" ELSE "d"."id" END AS "category_id_alias"
        # FROM "accounts" AS "a"
        # LEFT JOIN "categories" AS "b" ON (("a"."root" IS NULL) AND ("b"."id" = "a"."category_id"))
        # LEFT JOIN "accounts" AS "c" ON (("c"."root" IS NULL) AND ("c"."address" = "a"."root"))
        # LEFT JOIN "categories" AS "d" ON ("d"."id" = "c"."category_id")
        # LEFT JOIN "wallets" AS "e" ON ("e"."address" = "a"."address")

        # join the category table for root accounts
        sql_categories_join = sql_accounts_table.join(
            Table(CATEGORIES_TABLE_NAME), "LEFT"
        )
        sql_categories_join.condition = Equal(sql_accounts_table.root, None) & (
            sql_categories_join.right.id == sql_accounts_table.category_id
        )
        # join the account table to get root accounts of derived accounts
        sql_accounts_join = sql_categories_join.join(Table(TABLE_NAME), "LEFT")
        sql_accounts_join.condition = Equal(sql_accounts_join.right.root, None) & (
            sql_accounts_join.right.address == sql_accounts_table.root
        )
        # join the category table for root account of derived accounts
        sql_categories_join2 = sql_accounts_join.join(
            Table(CATEGORIES_TABLE_NAME), "LEFT"
        )
        sql_categories_join2.condition = (
            sql_categories_join2.right.id == sql_accounts_join.right.category_id
        )
        # join wallet table to get wallet
        sql_wallets_join = sql_categories_join2.join(Table(WALLETS_TABLE_NAME), "LEFT")
        sql_wallets_join.condition = (
            sql_wallets_join.right.address == sql_accounts_table.address
        )
        sql_identities_join = sql_wallets_join.join(
            Table(IDENTITIES_TABLE_NAME), "LEFT"
        )
        sql_identities_join.condition = (
            sql_identities_join.right.address == sql_accounts_table.address
        )

        # to sort by category name
        category_name_alias = Case(
            (Equal(sql_accounts_table.root, None), sql_categories_join.right.name),
            else_=sql_categories_join2.right.name,
        ).as_(CATEGORY_NAME_ALIAS_COLUMN)

        # to filter by category id
        category_id_alias = Case(
            (Equal(sql_accounts_table.root, None), sql_categories_join.right.id),
            else_=sql_categories_join2.right.id,
        ).as_(CATEGORY_ID_ALIAS_COLUMN)

        # create select query wrapper
        sql_select = sql_identities_join.select(
            sql_wallets_join.right.address,
            sql_identities_join.right.index_,
            sql_identities_join.right.name,
            sql_accounts_table.balance,
            sql_accounts_table.name,
            sql_accounts_table.address,
            sql_accounts_table.path,
            sql_accounts_table.root,
            sql_accounts_table.crypto_type,
            category_name_alias,
            category_id_alias,
        )

        sql_sort_columns = [
            sql_wallets_join.right.address,
            sql_identities_join.right.name,
            sql_accounts_table.name,
            sql_accounts_table.balance,
            sql_accounts_table.address,
            sql_accounts_table.path,
            sql_accounts_table.root,
            sql_accounts_table.crypto_type,
            category_name_alias,
        ]

        # if sort column...
        if sort_column_index is not None:
            # set sort column by its index
            sql_sort_column = sql_sort_columns[sort_column_index]
            if sql_sort_column is not None:
                sql_select.order_by = (
                    sql_sort_column.asc
                    if sort_order == AccountsRepositoryInterface.SORT_ORDER_ASCENDING
                    else sql_sort_column.desc
                )

        sql_filter_columns = {
            AccountsRepositoryInterface.TABLE_VIEW_FILTER_BY_CATEGORY_ID: sql_accounts_table.category_id.as_(
                CATEGORY_ID_ALIAS_COLUMN
            ),
            AccountsRepositoryInterface.TABLE_VIEW_FILTER_BY_WALLET: sql_wallets_join.right.address,
        }

        # create where conditions
        conditions = []
        if filters is not None:
            for key, value in filters.items():
                if key not in sql_filter_columns:
                    continue
                if key == AccountsRepositoryInterface.TABLE_VIEW_FILTER_BY_WALLET:
                    if value is True:
                        conditions.append(NotEqual(sql_filter_columns[key], None))
                    elif value is False:
                        conditions.append(Equal(sql_filter_columns[key], None))
                else:
                    conditions.append(sql_filter_columns[key] == value)

        # conditions are added with and operator
        def and_(a, b):
            return And((a, b))

        if len(conditions) > 0:
            sql_select.where = reduce(and_, conditions)  # type: ignore

        sql, args = tuple(sql_select)
        result_set = self.client.select(sql, args)

        list_ = []
        for row in result_set:
            table_view_row = get_table_view_row_from_row(
                list(
                    row[
                        : len(
                            AccountsRepositoryInterface.TableViewRow.__dataclass_fields__  # type: ignore
                        )
                    ]
                )
            )
            list_.append(table_view_row)
        return list_

    def add(self, account: Account) -> None:
        __doc__ = (  # pylint: disable=redefined-builtin, unused-variable
            AccountsRepositoryInterface.add.__doc__
        )

        # insert only non hidden fields
        self.client.insert(
            TABLE_NAME,
            **get_fields_from_account(account),
        )

    def update(self, account: Account) -> None:
        __doc__ = (  # pylint: disable=redefined-builtin, unused-variable
            AccountsRepositoryInterface.update.__doc__
        )

        # update only non hidden fields
        self.client.update(
            TABLE_NAME,
            f"address='{account.address}'",
            **get_fields_from_account(account),
        )

    def delete(self, address: str) -> None:
        __doc__ = (  # pylint: disable=redefined-builtin, unused-variable
            AccountsRepositoryInterface.delete.__doc__
        )

        self.client.delete(TABLE_NAME, address=address)

    def delete_all(self) -> None:
        __doc__ = (  # pylint: disable=redefined-builtin, unused-variable
            AccountsRepositoryInterface.delete_all.__doc__
        )

        self.client.clear(TABLE_NAME)

    def count(self) -> int:
        __doc__ = (  # pylint: disable=redefined-builtin, unused-variable
            AccountsRepositoryInterface.count.__doc__
        )
        row = self.client.select_one(f"SELECT count(address) FROM {TABLE_NAME}")
        if row is None:
            return 0

        return row[0]

    def total_balance(self) -> int:
        __doc__ = (  # pylint: disable=redefined-builtin, unused-variable
            AccountsRepositoryInterface.total_balance.__doc__
        )
        row = self.client.select_one(f"SELECT sum(balance) FROM {TABLE_NAME}")
        if row is None:
            return 0

        return row[0] or 0


def get_fields_from_account(account: Account) -> dict:
    """
    Return a dict of supported fields with normalized value

    :param account: Account instance
    :return:
    """
    fields = {}
    for (key, value) in account.__dict__.items():
        if key.startswith("_"):
            continue
        elif isinstance(value, UUID):
            fields[key] = value.hex
        elif key == "balance" and value is not None:
            # fix overflow with big numbers
            fields[key] = str(value)
        else:
            fields[key] = value

    return fields


def get_table_view_row_from_row(row: list) -> AccountsRepositoryInterface.TableViewRow:
    """
    Return AccountsRepositoryInterface.TableViewRow from DB result set row

    :param row: Result set row
    :return:
    """
    if row[3] is not None:
        # convert to integer from str type
        row[3] = int(row[3])
    return AccountsRepositoryInterface.TableViewRow(*row)


def get_account_from_row(row: tuple) -> Account:
    """
    Return an Account instance from a result set row

    :param row: Result set row
    :return:
    """
    values: List[Any] = []
    count = 0
    for value in row:
        if count == 3 and value is not None:
            # convert to integer from str type
            values.append(int(value))
        elif count == 6:
            values.append(bool(value))
        elif count == 7 and value is not None:
            values.append(UUID(hex=value))
        elif count == 9 and value is not None:
            # python datetime does not handle timezone until 3.11
            values.append(parser.parse(value))
        elif count == 10 and value is not None:
            # python datetime does not handle timezone until 3.11
            values.append(parser.parse(value))
        else:
            values.append(value)
        count += 1

    return Account(*values)  # pylint: disable=no-value-for-parameter



# Copyright 2021 Vincent Texier <vit@free.fr>
#
# This software is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
from __future__ import annotations

from tikka.adapters.repository.accounts import DBAccountsRepository
from tikka.adapters.repository.authorities import DBAuthoritiesRepository
from tikka.adapters.repository.categories import DBCategoriesRepository
from tikka.adapters.repository.config import FileConfigRepository
from tikka.adapters.repository.currencies import FileCurrenciesRepository
from tikka.adapters.repository.currency import DBCurrencyRepository
from tikka.adapters.repository.file_db_client import FileDBClient
from tikka.adapters.repository.file_wallets import V1FileWalletsRepository
from tikka.adapters.repository.identities import DBIdentitiesRepository
from tikka.adapters.repository.indexers import DBIndexersRepository
from tikka.adapters.repository.nodes import DBNodesRepository
from tikka.adapters.repository.passwords import DBPasswordsRepository
from tikka.adapters.repository.preferences import DBPreferencesRepository
from tikka.adapters.repository.smiths import DBSmithsRepository
from tikka.adapters.repository.technical_committee_members import (
    DBTechnicalCommitteeMembersRepository,
)
from tikka.adapters.repository.technical_committee_proposals import (
    DBTechnicalCommitteeProposalsRepository,
)
from tikka.adapters.repository.transfers import DBTransfersRepository
from tikka.adapters.repository.wallets import DBWalletsRepository
from tikka.interfaces.adapters.repository.repository import RepositoryInterface
from tikka.interfaces.adapters.repository.technical_committee_proposals import (
    TechnicalCommitteeProposalsRepositoryInterface,
)


class Repository(RepositoryInterface):
    """
    Repository class
    """

    def __init__(self, data_path: str) -> None:
        """
        Init Repository instance
        """
        # dynamic user config file
        self._config = FileConfigRepository(data_path)
        # static application config file
        self._currencies = FileCurrenciesRepository()

        # handle V1 file wallets
        self._v1_file_wallets = V1FileWalletsRepository()

        # database
        self._db_client = FileDBClient()
        self._accounts = DBAccountsRepository(self._db_client)
        self._authorities = DBAuthoritiesRepository(self._db_client)
        self._categories = DBCategoriesRepository(self._db_client)
        self._currency = DBCurrencyRepository(self._db_client)
        self._identities = DBIdentitiesRepository(self._db_client)
        self._indexers = DBIndexersRepository(self._db_client)
        self._nodes = DBNodesRepository(self._db_client)
        self._passwords = DBPasswordsRepository(self._db_client)
        self._preferences = DBPreferencesRepository(self._db_client)
        self._smiths = DBSmithsRepository(self._db_client)
        self._transfers = DBTransfersRepository(self._db_client)
        self._wallets = DBWalletsRepository(self._db_client)
        self._technical_committee_members = DBTechnicalCommitteeMembersRepository(
            self._db_client
        )
        self._technical_committee_proposals = DBTechnicalCommitteeProposalsRepository(
            self._db_client
        )

    @property
    def db_client(self) -> FileDBClient:
        __doc__ = (  # pylint: disable=redefined-builtin, unused-variable
            RepositoryInterface.db_client.__doc__
        )
        return self._db_client

    @property
    def config(self) -> FileConfigRepository:
        __doc__ = (  # pylint: disable=redefined-builtin, unused-variable
            RepositoryInterface.config.__doc__
        )
        return self._config

    @property
    def currencies(self) -> FileCurrenciesRepository:
        __doc__ = (  # pylint: disable=redefined-builtin, unused-variable
            RepositoryInterface.currencies.__doc__
        )
        return self._currencies

    @property
    def v1_file_wallets(self) -> V1FileWalletsRepository:
        __doc__ = (  # pylint: disable=redefined-builtin, unused-variable
            RepositoryInterface.v1_file_wallets.__doc__
        )
        return self._v1_file_wallets

    @property
    def accounts(self) -> DBAccountsRepository:
        __doc__ = (  # pylint: disable=redefined-builtin, unused-variable
            RepositoryInterface.accounts.__doc__
        )
        return self._accounts

    @property
    def authorities(self) -> DBAuthoritiesRepository:
        __doc__ = (  # pylint: disable=redefined-builtin, unused-variable
            RepositoryInterface.authorities.__doc__
        )
        return self._authorities

    @property
    def categories(self) -> DBCategoriesRepository:
        __doc__ = (  # pylint: disable=redefined-builtin, unused-variable
            RepositoryInterface.categories.__doc__
        )
        return self._categories

    @property
    def currency(self) -> DBCurrencyRepository:
        __doc__ = (  # pylint: disable=redefined-builtin, unused-variable
            RepositoryInterface.currency.__doc__
        )
        return self._currency

    @property
    def identities(self) -> DBIdentitiesRepository:
        __doc__ = (  # pylint: disable=redefined-builtin, unused-variable
            RepositoryInterface.identities.__doc__
        )
        return self._identities

    @property
    def indexers(self) -> DBIndexersRepository:
        __doc__ = (  # pylint: disable=redefined-builtin, unused-variable
            RepositoryInterface.indexers.__doc__
        )
        return self._indexers

    @property
    def nodes(self) -> DBNodesRepository:
        __doc__ = (  # pylint: disable=redefined-builtin, unused-variable
            RepositoryInterface.nodes.__doc__
        )
        return self._nodes

    @property
    def passwords(self) -> DBPasswordsRepository:
        __doc__ = (  # pylint: disable=redefined-builtin, unused-variable
            RepositoryInterface.passwords.__doc__
        )
        return self._passwords

    @property
    def preferences(self) -> DBPreferencesRepository:
        __doc__ = (  # pylint: disable=redefined-builtin, unused-variable
            RepositoryInterface.preferences.__doc__
        )
        return self._preferences

    @property
    def smiths(self) -> DBSmithsRepository:
        __doc__ = (  # pylint: disable=redefined-builtin, unused-variable
            RepositoryInterface.smiths.__doc__
        )
        return self._smiths

    @property
    def transfers(self) -> DBTransfersRepository:
        __doc__ = (  # pylint: disable=redefined-builtin, unused-variable
            RepositoryInterface.transfers.__doc__
        )
        return self._transfers

    @property
    def wallets(self) -> DBWalletsRepository:
        __doc__ = (  # pylint: disable=redefined-builtin, unused-variable
            RepositoryInterface.wallets.__doc__
        )
        return self._wallets

    @property
    def technical_committee_members(
        self,
    ) -> DBTechnicalCommitteeMembersRepository:
        __doc__ = (  # pylint: disable=redefined-builtin, unused-variable
            RepositoryInterface.technical_committee_members.__doc__
        )
        return self._technical_committee_members

    @property
    def technical_committee_proposals(
        self,
    ) -> TechnicalCommitteeProposalsRepositoryInterface:
        __doc__ = (  # pylint: disable=redefined-builtin, unused-variable
            RepositoryInterface.technical_committee_proposals.__doc__
        )
        return self._technical_committee_proposals



# Copyright 2021 Vincent Texier <vit@free.fr>
#
# This software is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
import logging

from tikka.domains.entities.currency import Currency
from tikka.interfaces.adapters.network.connection import NetworkConnectionError
from tikka.interfaces.adapters.network.node.currency import (
    NodeCurrencyException,
    NodeCurrencyInterface,
)


class NodeCurrency(NodeCurrencyInterface):
    """
    NodeCurrency class
    """

    def get_instance(self) -> Currency:
        __doc__ = (  # pylint: disable=redefined-builtin, unused-variable
            NodeCurrencyInterface.get_instance.__doc__
        )
        if (
            not self.node.connection.is_connected()
            or self.node.connection.client is None
        ):
            raise NodeCurrencyException(NetworkConnectionError())

        try:
            chain_name = self.node.connection.client.rpc_request(
                "system_chain", []
            ).get("result")
        except Exception as exception:
            logging.exception(exception)
            raise NodeCurrencyException(exception)

        try:
            result = self.node.connection.client.rpc_request(
                "system_properties", []
            ).get("result")
        except Exception as exception:
            logging.exception(exception)
            raise NodeCurrencyException(exception)

        token_decimals = result["tokenDecimals"]
        token_symbol = result["tokenSymbol"]

        try:
            multi_result = self.node.connection.client.query_multi(
                [
                    ("UniversalDividend", "CurrentUd", []),
                    ("UniversalDividend", "MonetaryMass", []),
                    ("Membership", "CounterForMembership", []),
                ]
            )
        except Exception as exception:
            logging.exception(exception)
            raise NodeCurrencyException(exception)

        current_ud = multi_result[0]
        monetary_mass = multi_result[1]
        counter_for_membership = multi_result[2]

        expected_block_time_in_seconds = (
            self.node.connection.client.get_constant("Babe", "ExpectedBlockTime") / 1000
        )
        epoch_duration_in_blocks = self.node.connection.client.get_constant(
            "Babe", "EpochDuration"
        )
        ss58_format = self.node.connection.client.get_constant("System", "SS58Prefix")
        system_infos = self.node.connection.client.get_constant("System", "Version")

        certification_number_to_be_member = self.node.connection.client.get_constant(
            "Wot", "MinCertForMembership"
        )

        minimum_delay_between_two_membership_renewals_in_blocks = (
            self.node.connection.client.get_constant(
                "Membership", "MembershipRenewalPeriod"
            )
        )
        validity_duration_of_membership_in_blocks = (
            self.node.connection.client.get_constant("Membership", "MembershipPeriod")
        )

        minimum_certifications_received_to_be_certifier = (
            self.node.connection.client.get_constant(
                "Certification", "MinReceivedCertToBeAbleToIssueCert"
            )
        )
        validity_duration_of_certification_in_blocks = (
            self.node.connection.client.get_constant("Certification", "ValidityPeriod")
        )
        minimum_delay_between_two_certifications_in_blocks = (
            self.node.connection.client.get_constant("Certification", "CertPeriod")
        )
        maximum_number_of_certifications_per_member = (
            self.node.connection.client.get_constant("Certification", "MaxByIssuer")
        )

        maximum_distance_in_step = self.node.connection.client.get_constant(
            "Distance", "MaxRefereeDistance"
        )
        minimum_percentage_of_remote_referral_members_to_be_member = (
            self.node.connection.client.get_constant(
                "Distance", "MinAccessibleReferees"
            )
            / 10000000
        )

        identity_automatic_revocation_period_in_blocks = (
            self.node.connection.client.get_constant("Identity", "AutorevocationPeriod")
        )
        minimum_delay_between_changing_identity_owner_in_blocks = (
            self.node.connection.client.get_constant("Identity", "ChangeOwnerKeyPeriod")
        )
        confirm_identity_period_in_blocks = self.node.connection.client.get_constant(
            "Identity", "ConfirmPeriod"
        )
        identity_deletion_after_revocation_in_blocks = (
            self.node.connection.client.get_constant("Identity", "DeletionPeriod")
        )
        minimum_delay_between_identity_creation_in_blocks = (
            self.node.connection.client.get_constant("Identity", "IdtyCreationPeriod")
        )
        identity_validation_period_in_blocks = self.node.connection.client.get_constant(
            "Identity", "ValidationPeriod"
        )
        maximum_certifications_per_smith = self.node.connection.client.get_constant(
            "SmithMembers", "MaxByIssuer"
        )
        number_of_certifications_to_become_smith = (
            self.node.connection.client.get_constant(
                "SmithMembers", "MinCertForMembership"
            )
        )
        maximum_inactivity_duration_allowed_for_smith_in_blocks = (
            self.node.connection.client.get_constant(
                "SmithMembers", "SmithInactivityMaxDuration"
            )
        )

        return Currency(
            code_name=system_infos["spec_name"],
            name=chain_name,
            ss58_format=ss58_format,
            token_decimals=token_decimals,
            token_symbol=token_symbol,
            universal_dividend=current_ud,
            monetary_mass=monetary_mass,
            members_count=counter_for_membership,
            block_duration=expected_block_time_in_seconds,
            epoch_duration=epoch_duration_in_blocks * expected_block_time_in_seconds,
            certification_number_to_be_member=certification_number_to_be_member,
            minimum_delay_between_two_membership_renewals=minimum_delay_between_two_membership_renewals_in_blocks
            * expected_block_time_in_seconds,
            validity_duration_of_membership=validity_duration_of_membership_in_blocks
            * expected_block_time_in_seconds,
            minimum_certifications_received_to_be_certifier=minimum_certifications_received_to_be_certifier,
            validity_duration_of_certification=validity_duration_of_certification_in_blocks
            * expected_block_time_in_seconds,
            minimum_delay_between_two_certifications=minimum_delay_between_two_certifications_in_blocks
            * expected_block_time_in_seconds,
            maximum_number_of_certifications_per_member=maximum_number_of_certifications_per_member,
            maximum_distance_in_step=maximum_distance_in_step,
            minimum_percentage_of_remote_referral_members_to_be_member=minimum_percentage_of_remote_referral_members_to_be_member,
            identity_automatic_revocation_period=identity_automatic_revocation_period_in_blocks
            * expected_block_time_in_seconds,
            minimum_delay_between_changing_identity_owner=minimum_delay_between_changing_identity_owner_in_blocks
            * expected_block_time_in_seconds,
            confirm_identity_period=confirm_identity_period_in_blocks
            * expected_block_time_in_seconds,
            identity_deletion_after_revocation=identity_deletion_after_revocation_in_blocks
            * expected_block_time_in_seconds,
            minimum_delay_between_identity_creation=minimum_delay_between_identity_creation_in_blocks
            * expected_block_time_in_seconds,
            identity_validation_period=identity_validation_period_in_blocks
            * expected_block_time_in_seconds,
            maximum_certifications_per_smith=maximum_certifications_per_smith,
            number_of_certifications_to_become_smith=number_of_certifications_to_become_smith,
            maximum_inactivity_duration_allowed_for_smith=maximum_inactivity_duration_allowed_for_smith_in_blocks
            * expected_block_time_in_seconds,
        )



# Copyright 2021 Vincent Texier <vit@free.fr>
#
# This software is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
import logging
import sys
from typing import Optional

from PyQt5.QtCore import QMutex, QSize
from PyQt5.QtGui import QFont, QMovie
from PyQt5.QtWidgets import QApplication, QDialog, QWidget, QWizard

from tikka.domains.application import Application
from tikka.domains.entities.account import AccountCryptoType
from tikka.domains.entities.constants import AMOUNT_UNIT_KEY, DATA_PATH
from tikka.interfaces.adapters.repository.accounts import AccountsRepositoryInterface
from tikka.slots.pyqt.entities.constants import (
    ADDRESS_MONOSPACE_FONT_NAME,
    ICON_LOADER,
    SELECTED_UNIT_PREFERENCES_KEY,
)
from tikka.slots.pyqt.entities.worker import AsyncQWorker
from tikka.slots.pyqt.resources.gui.windows.v1_account_import_wizard_rc import (
    Ui_importAccountV1Wizard,
)
from tikka.slots.pyqt.windows.account_unlock import AccountUnlockWindow


class V1AccountImportWizardWindow(QWizard, Ui_importAccountV1Wizard):
    """
    V1AccountImportWizardWindow class
    """

    def __init__(
        self, application: Application, mutex: QMutex, parent: Optional[QWidget] = None
    ):
        """
        Init import V1 account wizard window

        :param application: Application instance
        :param mutex: QMutex instance
        :param parent: QWidget instance
        """
        super().__init__(parent=parent)
        self.setupUi(self)

        self.application = application
        self._ = self.application.translator.gettext
        self.mutex = mutex

        unit_preference = self.application.repository.preferences.get(
            SELECTED_UNIT_PREFERENCES_KEY
        )
        if unit_preference is not None:
            self.amount = self.application.amounts.get_amount(unit_preference)
        else:
            self.amount = self.application.amounts.get_amount(AMOUNT_UNIT_KEY)

        # source account
        self.source_account = None

        # destination account
        self.destination_account = None
        self.destination_mnemonic = None
        self.destination_account = None
        self.destination_keypair = None

        # set monospace font to address fields
        self.monospace_font = QFont(ADDRESS_MONOSPACE_FONT_NAME)
        self.monospace_font.setStyleHint(QFont.Monospace)
        self.sourceV1AddressValueImportLabel.setFont(self.monospace_font)
        self.sourceAddressValueImportLabel.setFont(self.monospace_font)
        self.destinationAddressValueImportLabel.setFont(self.monospace_font)

        # animated loading icon
        self.loader_movie = QMovie(ICON_LOADER)
        self.loader_movie.setScaledSize(QSize(16, 16))
        self.loader_movie.start()

        ##############################
        # ASYNC METHODS
        ##############################
        # Create a QWorker object
        self.import_source_into_destination_on_network_async_qworker = AsyncQWorker(
            self.network_import_source_into_destination, self.mutex
        )
        self.wizard_page2_init()

    def wizard_page2_init(self):
        """
        Initialize page 2 form

        :return:
        """
        # fonts
        self.sourceV1AddressValueLabel.setFont(self.monospace_font)
        self.sourceAddressValueLabel.setFont(self.monospace_font)

        # page next button status handling
        self.wizardPage2.isComplete = self.wizard_page2_is_complete
        self.wizardPage2.validatePage = self.wizard_page2_validate_page

        # empty selection
        self.sourceV1AccountComboBox.addItem(
            "",
            userData=None,
        )
        # build selector
        for account in self.application.accounts.get_list(
            filters={
                AccountsRepositoryInterface.COLUMN_CRYPTO_TYPE: int(
                    AccountCryptoType.ED25519
                ),
            }
        ):
            # only accounts with password (user owns it)
            if account.balance is not None and self.application.passwords.exists(
                account.address
            ):
                account_label = (
                    account.name if account.name is not None else account.address
                )
                identity = self.application.identities.get_by_address(account.address)

                if identity is not None and identity.name is not None:
                    account_label = f"{identity.name0}#{identity.index}"
                self.sourceV1AccountComboBox.addItem(
                    account_label,
                    userData=account.address,
                )

        # events
        self.sourceV1AccountComboBox.currentIndexChanged.connect(
            self._source_generate_address
        )

    def wizard_page2_validate_page(self):
        """
        Ask password if source account unlocked to allow to go to next page

        :return:
        """
        if not self.application.wallets.is_unlocked(self.source_account.address):
            # ask password...
            dialog_code = AccountUnlockWindow(
                self.application, self.source_account, self
            ).exec_()
            if dialog_code == QDialog.Rejected:
                return False

        self.wizard_page3_init()
        return True

    def wizard_page2_is_complete(self) -> bool:
        """
        Function to overload V1AccountImportWizardWindow->Page2->IsComplete() method

        :return:
        """
        result = False

        if self.source_account is not None and self.source_account.balance is not None:
            result = True

        return result

    def _source_generate_address(self):
        """
        Generate address from ID

        :return:
        """
        self.sourceErrorLabel.setText("")

        selected_address = self.sourceV1AccountComboBox.currentData()
        if selected_address is None:
            self.sourceV1AddressValueLabel.setText("")
            self.sourceAddressValueLabel.setText("")
            self.sourceBalanceValueLabel.setText("?")
            self.sourceIdentityValueLabel.setText("")
            return

        self.source_account = self.application.accounts.get_by_address(
            self.sourceV1AccountComboBox.currentData()
        )

        # display v1 address
        self.sourceV1AddressValueLabel.setText(
            self.source_account.get_v1_address(
                self.application.currencies.get_current().ss58_format
            )
        )

        # if account name displayed in combobox...
        if (
            self.sourceV1AccountComboBox.currentText()
            != self.sourceV1AccountComboBox.currentData()
        ):
            self.sourceAddressValueLabel.setText(self.source_account.address)
        else:
            self.sourceAddressValueLabel.setText("")

        self.sourceBalanceValueLabel.setText(
            self.locale().toCurrencyString(
                self.amount.value(self.source_account.balance),
                self.amount.symbol(),
            )
        )
        identity = self.application.identities.get_by_address(
            self.source_account.address
        )
        if identity is not None:
            identity_name = identity.name or ""
            self.sourceIdentityValueLabel.setText(f"{identity_name}#{identity.index}")
        else:
            self.sourceIdentityValueLabel.setText(self._("None"))

        self.wizardPage2.completeChanged.emit()

    def wizard_page3_init(self):
        """
        Initialize page 3 form

        :return:
        """
        # fonts
        self.destinationAccountComboBox.setFont(self.monospace_font)
        self.destinationAddressValueLabel.setFont(self.monospace_font)

        # page next button status handling
        self.wizardPage3.isComplete = self.wizard_page3_is_complete
        self.wizardPage3.validatePage = self.wizard_page3_validate_page

        # init V2 accounts combo box
        identity_idx_source = self.application.identities.get_index_by_address(
            self.source_account.address
        )
        self.destinationAccountComboBox.addItem("", None)
        for account in self.application.accounts.get_list(
            filters={
                AccountsRepositoryInterface.COLUMN_CRYPTO_TYPE: int(
                    AccountCryptoType.SR25519
                ),
            }
        ):
            # we can not transfer a source identity on a destination account with already an identity
            identity_idx = self.application.identities.get_index_by_address(
                account.address
            )
            if identity_idx_source is not None and identity_idx is not None:
                continue
            # only accounts with password (user owns it)
            if self.application.passwords.exists(account.address):
                self.destinationAccountComboBox.addItem(
                    account.name if account.name is not None else account.address,
                    userData=account.address,
                )

        # events
        self.destinationAccountComboBox.currentIndexChanged.connect(
            self._generate_destination_address
        )

    def wizard_page3_is_complete(self) -> bool:
        """
        Function to overload V1AccountImportWizardWindow->Page3->IsComplete() method

        :return:
        """
        result = False

        if self.destination_account is not None and (
            self.application.identities.get_index_by_address(
                self.source_account.address
            )
            is None
            or (
                self.application.identities.get_index_by_address(
                    self.source_account.address
                )
                is not None
                and self.application.identities.get_index_by_address(
                    self.destination_account.address
                )
                is None
            )
        ):
            result = True

        return result

    def wizard_page3_validate_page(self):
        """
        Ask password if destination account unlocked to allow to go to next page

        :return:
        """
        if not self.application.wallets.is_unlocked(self.destination_account.address):
            # ask password...
            dialog_code = AccountUnlockWindow(
                self.application, self.destination_account, self
            ).exec_()
            if dialog_code == QDialog.Rejected:
                return False

        self.wizard_page4_init()
        return True

    def _generate_destination_address(self):
        """
        Generate destination address

        :return:
        """
        self.destination_account = self.application.accounts.get_by_address(
            self.destinationAccountComboBox.currentData()
        )
        self.destinationErrorLabel.setText("")

        if self.destination_account is None:
            self.destinationDerivationValueLabel.setText("")
            self.destinationAddressValueLabel.setText("")
            self.destinationBalanceValueLabel.setText("?")
            return

        self.destinationDerivationValueLabel.setText(self.destination_account.path)
        self.destinationAddressValueLabel.setText(self.destination_account.address)
        if self.destination_account.balance is not None:
            self.sourceBalanceValueLabel.setText(
                self.locale().toCurrencyString(
                    self.amount.value(self.source_account.balance),
                    self.amount.symbol(),
                )
            )
        else:
            self.destinationBalanceValueLabel.setText("?")

        self.wizardPage3.completeChanged.emit()

    def wizard_page4_init(self):
        """
        Initialize page 4 form

        :return:
        """
        # fonts
        self.sourceV1AddressValueImportLabel.setFont(self.monospace_font)
        self.sourceAddressValueImportLabel.setFont(self.monospace_font)
        self.destinationAddressValueImportLabel.setFont(self.monospace_font)

        self.sourceV1AddressValueImportLabel.setText(
            self.source_account.get_v1_address(
                self.application.currencies.get_current().ss58_format
            )
        )
        self.sourceAddressValueImportLabel.setText(self.source_account.address)
        if self.source_account is not None:
            if self.source_account.balance is not None:
                self.sourceBalanceValueImportLabel.setText(
                    self.locale().toCurrencyString(
                        self.amount.value(self.source_account.balance),
                        self.amount.symbol(),
                    )
                )
            identity = self.application.identities.get_by_address(
                self.source_account.address
            )
            if identity is not None:
                identity_name = identity.name or ""
                self.sourceIdentityValueImportLabel.setText(
                    f"{identity_name}#{identity.index}"
                )
            else:
                self.sourceIdentityValueImportLabel.setText(self._("None"))
        else:
            self.sourceIdentityValueImportLabel.setText(self._("None"))

        self.destinationAddressValueImportLabel.setText(
            self.destination_account.address
        )
        if self.destination_account.balance is not None:
            self.destinationBalanceValueImportLabel.setText(
                self.locale().toCurrencyString(
                    self.amount.value(self.destination_account.balance),
                    self.amount.symbol(),
                )
            )
        self.destinationIdentityValueImportLabel.setText(self._("None"))

        # page next button status handling
        self.wizardPage4.isComplete = self.wizard_page4_is_complete

        # events
        self.importButton.clicked.connect(self.on_import_button_clicked)

        ##############################
        # ASYNC METHODS
        ##############################
        self.import_source_into_destination_on_network_async_qworker.finished.connect(
            self._on_finished_import_source_into_destination_on_network
        )

    def wizard_page4_is_complete(self) -> bool:
        """
        Function to overload V1AccountImportWizardWindow->Page4->IsComplete() method

        :return:
        """
        result = False

        if (
            self.source_account is not None
            and (
                self.source_account.balance == 0 or self.source_account.balance is None
            )
            and self.application.identities.get_index_by_address(
                self.source_account.address
            )
            is None
        ):
            result = True

        return result

    def on_import_button_clicked(self):
        """
        Triggered when user click on import button

        :return:
        """
        self.importButton.setDisabled(True)
        self.importErrorLabel.setMovie(self.loader_movie)

        self.import_source_into_destination_on_network_async_qworker.start()

    def network_import_source_into_destination(self):
        """
        Send changeOwnerKey for identity if any and tranfer money from V1 source account to V2 destination account

        :return:
        """
        self.importErrorLabel.setText("")

        source_keypair = self.application.wallets.get_keypair(
            self.source_account.address
        )
        destination_keypair = self.application.wallets.get_keypair(
            self.destination_account.address
        )

        if (
            self.application.identities.get_index_by_address(
                self.source_account.address
            )
            is not None
        ):
            try:
                self.application.identities.network_change_owner_key(
                    source_keypair, destination_keypair
                )
            except Exception as exception:
                self.importErrorLabel.setText(self._(str(exception)))
                # do not transfer all the money if identity transfer failed,
                # because identity transfer requires fees from sender account
                logging.exception(exception)
                return

        if self.source_account.balance is not None and self.source_account.balance > 0:
            try:
                self.application.transfers.network_send_all(
                    source_keypair, self.destination_account.address, False
                )
            except Exception as exception:
                self.importErrorLabel.setText(self._(str(exception)))
                logging.exception(exception)
                return
        try:
            self.source_account = self.application.accounts.network_update_balance(
                self.source_account
            )
        except Exception as exception:
            self.importErrorLabel.setText(self._(str(exception)))
            logging.exception(exception)
            return

        try:
            self.application.identities.network_update_identity(
                self.source_account.address
            )
        except Exception as exception:
            self.importErrorLabel.setText(self._(str(exception)))
            logging.exception(exception)
            return

        try:
            self.destination_account = self.application.accounts.network_update_balance(
                self.destination_account
            )
        except Exception as exception:
            self.importErrorLabel.setText(self._(str(exception)))
            logging.exception(exception)
            return

        try:
            self.application.identities.network_update_identity(
                self.destination_account.address
            )
        except Exception as exception:
            self.importErrorLabel.setText(self._(str(exception)))
            logging.exception(exception)

    def _on_finished_import_source_into_destination_on_network(self):
        """
        Triggered when async request import_source_into_destination_on_network is finished

        :return:
        """
        self.importErrorLabel.setMovie(None)

        if self.source_account is not None:
            if self.source_account.balance is not None:
                self.sourceBalanceValueImportLabel.setText(
                    self.locale().toCurrencyString(
                        self.amount.value(self.source_account.balance),
                        self.amount.symbol(),
                    )
                )
            else:
                # empty account deleted from blockchain
                self.sourceBalanceValueImportLabel.setText("?")

        identity_index = self.application.identities.get_index_by_address(
            self.source_account.address
        )
        if identity_index is not None:
            identity = self.application.identities.get(identity_index)
            identity_name = identity.name or ""
            self.sourceIdentityValueImportLabel.setText(
                f"{identity_name}#{identity.index}"
            )
        else:
            self.sourceIdentityValueImportLabel.setText(self._("None"))

        if self.destination_account.balance is not None:
            self.destinationBalanceValueImportLabel.setText(
                self.locale().toCurrencyString(
                    self.amount.value(self.destination_account.balance),
                    self.amount.symbol(),
                )
            )
        else:
            self.destinationBalanceValueImportLabel.setText("?")

        identity_index = self.application.identities.get_index_by_address(
            self.destination_account.address
        )
        if identity_index is not None:
            identity = self.application.identities.get(identity_index)
            identity_name = identity.name or ""
            self.destinationIdentityValueImportLabel.setText(
                f"{identity_name}#{identity.index}"
            )
        else:
            self.destinationIdentityValueImportLabel.setText(self._("None"))

        self.wizardPage4.completeChanged.emit()

        if self.wizardPage4.isComplete():
            self.importButton.setDisabled(True)
            self.importErrorLabel.setStyleSheet("color: green;")
            self.importErrorLabel.setText(self._("Account imported successfully!"))
            if (
                self.application.accounts.get_by_address(
                    self.destination_account.address
                )
                is None
            ):
                self.application.accounts.add(self.destination_account)
            else:
                self.application.accounts.update(self.destination_account)
        else:
            self.importButton.setDisabled(False)
            self.importErrorLabel.setStyleSheet("color: red;")
            if (
                self.source_account.balance is not None
                and self.source_account.balance > 0
            ):
                self.importErrorLabel.setText(self._("Error importing money!"))
            elif (
                self.application.identities.get_index_by_address(
                    self.destination_account.address
                )
                is None
            ):
                self.importErrorLabel.setText(self._("Error importing identity!"))


if __name__ == "__main__":
    qapp = QApplication(sys.argv)
    application_ = Application(DATA_PATH)
    V1AccountImportWizardWindow(application_, QMutex()).exec_()



# Copyright 2021 Vincent Texier <vit@free.fr>
#
# This software is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
from tikka.domains.accounts import Accounts
from tikka.domains.currencies import Currencies
from tikka.domains.entities.account import Account, AccountCryptoType
from tikka.domains.entities.constants import DERIVATION_SCAN_MAX_NUMBER
from tikka.interfaces.adapters.network.node.accounts import NodeAccountsInterface
from tikka.libs.keypair import Keypair


class Vaults:
    """
    Vaults domain class
    """

    def __init__(
        self,
        network: NodeAccountsInterface,
        accounts: Accounts,
        currencies: Currencies,
    ):
        """
        Init Vaults domain

        :param network: NetworkAccountsInterface instance
        :param accounts: Accounts domain instance
        :param currencies: Currencies domain instance
        """
        self.network = network
        self.accounts = accounts
        self.currencies = currencies

    def import_from_network(
        self, mnemonic: str, language_code: str, name: str, password: str
    ) -> None:
        """
        Return an Account instance from arguments

        :param mnemonic: Mnemonic phrase
        :param language_code: Mnemonic language code
        :param name: Root account name
        :param password: Vault/Root account password

        :return:
        """
        root_keypair = Keypair.create_from_uri(
            mnemonic,
            language_code=language_code,
            ss58_format=self.currencies.get_current().ss58_format,
            crypto_type=AccountCryptoType.SR25519,
        )
        root_account = self.accounts.get_by_address(root_keypair.ss58_address)
        if root_account is None:
            root_account = self.accounts.create_new_root_account(
                mnemonic,
                language_code,
                name,
                password,
                add_event=False,
            )

        self.import_derived_account_from_network(
            root_account, mnemonic, language_code, name, password  # type: ignore
        )

    def import_derived_account_from_network(
        self,
        root_account: Account,
        mnemonic: str,
        language_code: str,
        name: str,
        password: str,
    ):
        """
        Import derived account from network for root_account

        :param root_account: Root Account instance
        :param mnemonic: Mnemonic phrase
        :param language_code: Mnemonic language code
        :param name: Root account name
        :param password: Vault/Root account password

        :return:
        """
        derived_accounts = []
        addresses = []
        for derivation_number in range(0, DERIVATION_SCAN_MAX_NUMBER + 1):
            derivation = f"//{derivation_number}"
            suri = mnemonic + derivation
            keypair = Keypair.create_from_uri(
                suri,
                language_code=language_code,
                ss58_format=self.currencies.get_current().ss58_format,
                crypto_type=AccountCryptoType.SR25519,
            )
            derived_account = self.accounts.get_instance(
                keypair.ss58_address, f"{name}{derivation}"
            )
            derived_account.path = derivation
            derived_accounts.append(derived_account)
            addresses.append(keypair.ss58_address)

        balances = self.accounts.network_get_balances(addresses)

        for derived_account in derived_accounts:
            if balances[derived_account.address] is not None:
                existing_account = self.accounts.get_by_address(derived_account.address)
                if existing_account is not None:
                    existing_account.root = root_account.address
                    self.accounts.repository.update(existing_account)
                else:
                    balance = balances[derived_account.address]
                    if balance is not None:
                        account = self.accounts.create_new_account(
                            mnemonic,
                            language_code,
                            derived_account.path,  # type: ignore
                            derived_account.name,  # type: ignore
                            password,
                        )
                        account.balance = balance
                        self.accounts.repository.update(account)
            else:
                self.accounts.repository.delete(derived_account.address)



# Copyright 2021 Vincent Texier <vit@free.fr>
#
# This software is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from typing import Optional

from tikka.domains.entities.currency import Currency
from tikka.interfaces.adapters.repository.currency import CurrencyRepositoryInterface
from tikka.interfaces.adapters.repository.db_repository import DBRepositoryInterface

TABLE_NAME = "currency"


class DBCurrencyRepository(CurrencyRepositoryInterface, DBRepositoryInterface):
    """
    DBCurrencyRepository class
    """

    def add(self, currency: Currency) -> None:
        __doc__ = (  # pylint: disable=redefined-builtin, unused-variable
            DBCurrencyRepository.add.__doc__
        )

        # insert only non hidden fields
        self.client.insert(
            TABLE_NAME,
            **get_fields_from_currency(currency),
        )

    def get(self, code_name: str) -> Optional[Currency]:
        __doc__ = (  # pylint: disable=redefined-builtin, unused-variable
            CurrencyRepositoryInterface.get.__doc__
        )

        row = self.client.select_one(
            f"SELECT * FROM {TABLE_NAME} WHERE code_name=?", (code_name,)
        )

        if row is None:
            return None

        return Currency(*row)

    def update(self, currency: Currency) -> None:
        __doc__ = (  # pylint: disable=redefined-builtin, unused-variable
            CurrencyRepositoryInterface.update.__doc__
        )
        # update only non hidden fields
        self.client.update(
            TABLE_NAME,
            f"code_name='{currency.code_name}'",
            **get_fields_from_currency(currency),
        )


def get_fields_from_currency(currency: Currency) -> dict:
    """
    Return a dict of supported fields with normalized value

    :param currency: Currency instance
    :return:
    """
    fields = {}
    for (key, value) in currency.__dict__.items():
        if key.startswith("_"):
            continue
        elif key == "universal_dividend" and value is not None:
            # fix overflow with big numbers
            fields[key] = str(value)
        elif key == "monetary_mass" and value is not None:
            # fix overflow with big numbers
            fields[key] = str(value)
        else:
            fields[key] = value

    return fields


def get_currency_from_row(row: tuple) -> Currency:
    """
    Return a Currency instance from a result set row

    :param row: Result set row
    :return:
    """
    values = list(row)
    # # convert to integer from str type
    if values[5] is not None:
        values[5] = int(values[5])
    if values[6] is not None:
        values[6] = int(values[6])
    return Currency(*values)
